##
# File:    OeSearchMoleculeProvider.py
# Author:  J. Westbrook
# Date:    4-Mar-2020
#
# Updates:
#
##
"""
Utilities deliver OE molecule data for searchable chemical component data.
"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import logging
import os
import time

from rcsb.utils.chem.ChemCompSearchIndexProvider import ChemCompSearchIndexProvider
from rcsb.utils.chem.OeIoUtils import OeIoUtils
from rcsb.utils.io.MarshalUtil import MarshalUtil

# from rcsb.utils.io.SingletonClass import SingletonClass

logger = logging.getLogger(__name__)


class OeSearchMoleculeProvider(object):
    """Utilities build and deliver OE molecules for search applications. Source molecular
    definitions are taken from SMILES descriptors generated by ChemCompSearchIndexProvider()
    """

    def __init__(self, **kwargs):
        """Utilities build and deliver OE molecules for search applications. Source molecular
           definitions are taken from SMILES descriptors generated by ChemCompSearchIndexProvider()

        Args:
            cachePath (str, optional): path to the directory containing cache files (default: '.')
            ccFileNamePrefix (str, optional) file name prefix for chemical component search index (default: "cc")
            oeFileNamePrefix (str, optional) file name prefix for all generated databases (default: "oe")

        """
        # Database file names with be prefixed with base prefix plus the molecular build type and perception options
        oeFileNamePrefixBase = kwargs.get("oeFileNamePrefix", "oe")
        self.__ccFileNamePrefix = kwargs.get("ccFileNamePrefix", "oe")
        limitPerceptions = kwargs.get("limitPerceptions", False)
        if limitPerceptions:
            self.__oeFileNamePrefix = oeFileNamePrefixBase + "-limit"
        else:
            self.__oeFileNamePrefix = oeFileNamePrefixBase
        #
        cachePath = kwargs.get("cachePath", ".")
        self.__dirPath = os.path.join(cachePath, "oe_mol")
        #
        self.__fpDbD = {}
        self.__ssDb = None
        self.__oeMolD = {}
        self.__oeMolDb = None
        self.__oeMolDbTitleD = None
        #
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        self.__reload(**kwargs)

    def testCache(self):
        return self.__mU.exists(os.path.join(self.__dirPath, self.__getOeSearchMolFileName())) and self.__mU.exists(os.path.join(self.__dirPath, self.__getOeMolDbFileName()))

    def getSubSearchDb(self, screenType="SMARTS", numProc=1, forceRefresh=False):
        if not self.__ssDb or forceRefresh:
            oeIo = OeIoUtils()
            fp = os.path.join(self.__dirPath, self.__getSubSearchFileName(screenType))
            logger.info("Opening screened substructure search database %r", fp)
            self.__ssDb = oeIo.loadOeSubSearchDatabase(fp, screenType, numProc=numProc)
        return self.__ssDb

    def getFingerPrintDb(self, fpType, fpDbType="STANDARD", rebuild=False):
        if fpType not in self.__fpDbD or rebuild:
            oeIo = OeIoUtils()
            fastFpDbPath = os.path.join(self.__dirPath, self.__getFastFpDbFileName(fpType))
            oeMolDbFilePath = os.path.join(self.__dirPath, self.__getOeMolDbFileName())
            fpDb = oeIo.loadOeFingerPrintDatabase(oeMolDbFilePath, fastFpDbPath, inMemory=True, fpType=fpType, fpDbType=fpDbType)
            if fpDb:
                self.__fpDbD[fpType] = fpDb
        #
        return self.__fpDbD[fpType]

    def __getOeMolDbTitleIndex(self):
        oeMolDbTitleD = {}
        try:
            for idx in range(self.__oeMolDb.GetMaxMolIdx()):
                oeMolDbTitleD[self.__oeMolDb.GetTitle(idx)] = idx
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return oeMolDbTitleD

    def getOeMolDatabase(self):
        if not self.__oeMolDb:

            oeIo = OeIoUtils()
            self.__oeMolDb = oeIo.loadOeBinaryDatabaseAndIndex(os.path.join(self.__dirPath, self.__getOeMolDbFileName()))
            self.__oeMolDbTitleD = self.__getOeMolDbTitleIndex()
        return self.__oeMolDb, self.__oeMolDbTitleD

    def getOeMolD(self):
        try:
            if not self.__oeMolD:
                startTime = time.time()
                oeIo = OeIoUtils()
                self.__oeMolD = oeIo.readOeBinaryMolCache(os.path.join(self.__dirPath, self.__getOeSearchMolFileName()))
                logger.info("Loading OE binary molecule cache length %d (%.4f seconds)", len(self.__oeMolD), time.time() - startTime)
            return self.__oeMolD
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return None

    def getMol(self, searchCcId):
        try:
            if not self.__oeMolD:
                startTime = time.time()
                oeIo = OeIoUtils()
                self.__oeMolD = oeIo.readOeBinaryMolCache(os.path.join(self.__dirPath, self.__getOeSearchMolFileName()))
                logger.info("Loading OE binary molecule cache length %d (%.4f seconds)", len(self.__oeMolD), time.time() - startTime)
            return self.__oeMolD[searchCcId] if searchCcId in self.__oeMolD else None
        except Exception as e:
            logger.exception("Get molecule %r failing with %s", searchCcId, str(e))
        return None

    def __getFastFpDbFileName(self, fpType):
        return "%s-si-fast-fp-database-%s.fpbin" % (self.__oeFileNamePrefix, fpType)

    def __getSubSearchFileName(self, screenType):
        return "%s-si-ss-database-%s.oeb" % (self.__oeFileNamePrefix, screenType)

    def __getOeMolDbFileName(self):
        return "%s-si-mol-db-components.oeb" % self.__oeFileNamePrefix

    def __getOeSearchMolFileName(self):
        """Raw binary files of OE molecules in the search index.

        Returns:
            str: file name
        """
        return "%s-si-search-mol-components.oeb" % self.__oeFileNamePrefix

    def __reload(self, **kwargs):
        """Reload the dictionary of OE molecules and related data artifacts for chemical component definitions.

        Args:
            limitPerceptions(bool): process input descriptors in essentially verbatim mode (default: True)
            fpTypeList (list): fingerprint type (TREE,PATH,MACCS,CIRCULAR,LINGO)
            screenTypeList (list): fast sub search screen type (MOLECULE, SMARTS, MDL, ... )
            useCache (bool, optional): flag to use cached files. Defaults to True.
            cachePath (str): path to the top cache directory. Defaults to '.'.
            numProc (int): number processors to engage in screen substructure search database generation.
            suppressHydrogens (bool, optional): flag to suppress explicit hydrogens in the OE data store.
            molLimit (int):

        Returns:
            (bool) : True for success or False othewise

        """
        try:
            useCache = kwargs.get("useCache", True)
            cachePath = kwargs.get("cachePath", ".")
            numProc = kwargs.get("numProc", 2)
            molLimit = kwargs.get("molLimit", None)
            fpTypeList = kwargs.get("fpTypeList", ["TREE", "PATH", "MACCS", "CIRCULAR", "LINGO"])
            # screenTypeList = kwargs.get("screenTypeList", ["SMARTS"])
            screenTypeList = kwargs.get("screenTypeList", None)

            limitPerceptions = kwargs.get("limitPerceptions", False)
            suppressHydrogens = kwargs.get("suppressHydrogens", False)
            quietFlag = kwargs.get("quietFlag", True)
            logSizes = kwargs.get("logSizes", False)
            fpDbType = "STANDARD"
            buildScreenedDb = True
            #
            oeCount = 0
            errCount = 0
            failIdList = []
            oeIo = OeIoUtils(quietFlag=quietFlag)
            # --------
            oeSearchMolFilePath = os.path.join(self.__dirPath, self.__getOeSearchMolFileName())
            if not useCache or (useCache and not self.__mU.exists(oeSearchMolFilePath)):
                cmpKwargs = {k: v for k, v in kwargs.items() if k not in ["cachePath", "useCache", "molLimit"]}
                ccsiP = ChemCompSearchIndexProvider(cachePath=cachePath, useCache=True, molLimit=molLimit, **cmpKwargs)
                ok = ccsiP.testCache(minCount=molLimit, logSizes=logSizes)
                # ----
                ccIdxD = ccsiP.getIndex() if ok else {}
                idxCount = len(ccIdxD)
                # ------- JDW OE mol construction here -----
                startTime = time.time()
                oeCount, errCount, failIdList = oeIo.buildOeBinaryMolCacheFromIndex(
                    oeSearchMolFilePath, ccIdxD, quietFlag=quietFlag, fpTypeList=fpTypeList, limitPerceptions=limitPerceptions, suppressHydrogens=suppressHydrogens
                )
                if failIdList:
                    logger.info("failures %r", failIdList)
                endTime = time.time()
                logger.info("Constructed %d/%d cached oeMols  (unconverted %d) (%.4f seconds)", oeCount, idxCount, errCount, endTime - startTime)
            # --------
            oeMolDbFilePath = os.path.join(self.__dirPath, self.__getOeMolDbFileName())
            if not useCache or (useCache and not self.__mU.exists(oeMolDbFilePath)):
                startTime = time.time()
                molCount = oeIo.createOeBinaryDatabaseAndIndex(oeSearchMolFilePath, oeMolDbFilePath)
                endTime = time.time()
                logger.info("Created and stored %d indexed oeMols in OE database format (%.4f seconds)", molCount, endTime - startTime)

            # --------
            if fpDbType == "FAST":
                for fpType in fpTypeList:
                    startTime = time.time()
                    #  Fast FP search database file names
                    fpPath = os.path.join(self.__dirPath, self.__getFastFpDbFileName(fpType))
                    if not useCache or (useCache and not self.__mU.exists(fpPath)):
                        ok = oeIo.createOeFingerPrintDatabase(oeMolDbFilePath, fpPath, fpType=fpType)
                        endTime = time.time()
                        logger.info("Created and stored %s fingerprint database (%.4f seconds)", fpType, endTime - startTime)
            # --------
            if buildScreenedDb and screenTypeList:
                for screenType in screenTypeList:
                    startTime = time.time()
                    fp = os.path.join(self.__dirPath, self.__getSubSearchFileName(screenType))
                    if not useCache or (useCache and not self.__mU.exists(fp)):
                        ok = oeIo.createOeSubSearchDatabase(oeSearchMolFilePath, fp, screenType=screenType, numProc=numProc)
                        endTime = time.time()
                        logger.info("Constructed screened substructure database (status %r) with screenType %s (%.4f seconds)", ok, screenType, endTime - startTime)
                        # ---------
                        ssDb = oeIo.loadOeSubSearchDatabase(fp, screenType=screenType, numProc=numProc)
                        ok = ssDb.NumMolecules() == oeCount
                        # ----------
            #
            return True
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return False



##
# File:  OeAlignUtils.py
# Date:  17-Dec-2020  J. Westbrook
#
# Updates:
#  22-Jan-2021 jdw rename to OeAlignUtils() and add substructure mode
##
"""
Utilities for performing substructure and maximum common substructure molecular alignments.
"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"


import logging
import os
from collections import namedtuple

from openeye import oechem

from rcsb.utils.chem.OeCommonUtils import OeCommonUtils
from rcsb.utils.chem.OeIoUtils import OeIoUtils
from rcsb.utils.chem.OeMoleculeFactory import OeMoleculeFactory
from rcsb.utils.io.decorators import timeout
from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)

ComponentAtomDetails = namedtuple("ComponentAtomDetails", "atIdx atNo atName atType x y z atFormalCharge")
AlignAtomMap = namedtuple("AlignAtomMap", "refId refAtIdx refAtNo refAtName fitId fitAtIdx fitAtNo fitAtName")
AlignAtomUnMapped = namedtuple("AlignAtomUnMapped", "fitId fitAtIdx fitAtNo fitAtType fitAtName fitAtFormalCharge x y z fitNeighbors")


class OeAlignUtils(object):
    """Perform substructure and maximum common substructure molecular alignments.  Targets can be chemical component
    identifiers or paths to chemical component definition files.  Inputs can be in the the form of pairs,
    lists, and pair lists of chemical component definitions.

    """

    def __init__(self, verbose=True):
        #
        self.__verbose = verbose
        #
        self.__refId = None
        self.__refmol = None
        self.__refTitle = None
        #
        self.__fitId = None
        self.__fitmol = None
        self.__fitTitle = None
        #
        self.__pairTupleList = []
        #
        self.__searchType = "relaxed"
        #
        self.__refFD = {}
        self.__fitFD = {}
        #
        self.__mcss = None
        self.__ss = None
        self.__refPath = None
        self.__fitPath = None

    def setSearchType(self, sType="relaxed"):
        self.__searchType = sType
        return self.__searchType

    def setRefId(self, ccId, title=None, suppressHydrogens=False, cachePath="/data/components/ligand-dict-v3"):
        """Set the query molecule for MCSS comparison using the input chemical component ID.
        It is assumed that the definition for this ID can be obtained from the chemical component
        repository.

        Once the reference molecule is built, the MCSS calculation is initialized.

        A title is optionally provided otherwise the component Id will be used.

        The hydrogen flag can be used to perform the MCSS using only heavy atoms.
        """
        self.__refId = ccId
        ccIdU = ccId.upper()
        self.__refPath = os.path.join(cachePath, ccIdU[0], ccIdU, ccIdU + ".cif")

        _, self.__refmol, self.__refFD = self.getCCDefFile(self.__refPath, suppressHydrogens=suppressHydrogens)
        #
        if title is not None:
            self.__refmol.SetTitle(title)
            self.__refTitle = title
        else:
            self.__refmol.SetTitle(self.__refId)
            self.__refTitle = None
        # self.__setupMCSS(self.__refmol)

    def setRefPath(self, ccPath, title=None, suppressHydrogens=False, fType="CC", importType="2D"):
        """Set the query molecule for MCSS comparison using the input file path.

        The file type is either ['CC'] for a chemical component definition or another file type
        supported by OE toolkit assumed to have a conventional file extension for this type.

        Once the reference molecule is built, the MCSS calculation is initialized.

        A title is optionally provided otherwise the component Id will be used.

        The hydrogen flag can be used to perform the MCSS using only heavy atoms.
        """
        self.__refPath = ccPath
        if fType in ["CC"]:
            (self.__refId, self.__refmol, self.__refFD) = self.getCCDefFile(ccPath, suppressHydrogens=suppressHydrogens)
        else:
            (self.__refId, self.__refmol, self.__refFD) = self.__getMiscFile(ccPath, suppressHydrogens=suppressHydrogens, importType=importType, title=title)

        if self.__verbose:
            logger.debug("Derived ref ID     = %s", self.__refId)
            logger.debug("SMILES (stereo)  = %s", self.__refFD["SMILES_STEREO"])
        #
        # Insert title here -
        if title is not None:
            self.__refmol.SetTitle(title)
            self.__refTitle = title
        else:
            self.__refmol.SetTitle(self.__refId)
            self.__refTitle = None

        # self.__setupMCSS(self.__refmol)

    def setFitId(self, ccId, title=None, suppressHydrogens=False, cachePath="/data/components/ligand-dict-v3"):
        """Set the ID of the target/library molecule for MCSS comparison."""
        self.__fitId = ccId
        ccIdU = ccId.upper()
        self.__fitPath = os.path.join(cachePath, ccIdU[0], ccIdU, ccIdU + ".cif")
        self.__fitId, self.__fitmol, self.__fitFD = self.getCCDefFile(self.__fitPath, suppressHydrogens=suppressHydrogens)
        if self.__verbose:
            logger.info("Fit ID             = %s", self.__fitId)
            logger.info("SMILES (isomeric)  = %s", self.__fitFD["SMILES_STEREO"])
        if title is not None:
            self.__fitmol.SetTitle(title)
            self.__fitTitle = title
        else:
            self.__fitmol.SetTitle(self.__fitId)
            self.__fitTitle = None

    def setFitPath(self, ccPath, title=None, suppressHydrogens=False, fType="CC", importType="2D"):
        """Set the path to the target/library molecule for MCSS comparison using the input file path.

        The file type is either 'CC' for a chemical component definition or another file type
        supported by OE toolkit assumed to have a conventional file extension for this type.

        A title is optionally provided otherwise the component Id will be used.

        The hydrogen flag can be used to perform the MCSS using only heavy atoms.
        """
        self.__fitPath = ccPath
        if fType in ["CC"]:
            (self.__fitId, self.__fitmol, self.__fitFD) = self.getCCDefFile(ccPath, suppressHydrogens=suppressHydrogens)
        else:
            (self.__fitId, self.__fitmol, self.__fitFD) = self.__getMiscFile(ccPath, suppressHydrogens=suppressHydrogens, importType=importType, title=title)

        if self.__verbose:
            logger.debug("Derived fit ID     = %s", self.__fitId)
            logger.debug("SMILES (stereo)  = %s", self.__fitFD["SMILES_STEREO"])
        #
        # Insert title here -
        if title is not None:
            self.__fitmol.SetTitle(title)
            self.__fitTitle = title
        else:
            self.__fitmol.SetTitle(self.__refId)
            self.__fitTitle = None

    def __setFitIdList(self, ccIdList, cachePath="/data/components/ligand-dict-v3"):
        """Set the list of IDs to be compared with reference molecule by MCSS.

        From the input ID list build the internal pair list of
        tuples  [(refId,refPath,refTitle,fitId,fitPath,fitTitle),(),...]
        """
        self.__pairTupleList = []
        for ccId in ccIdList:
            refId = self.__refId
            refPath = self.__refPath
            #
            fitId = ccId
            ccIdU = ccId.upper()
            fitPath = os.path.join(cachePath, ccIdU[0], ccIdU, ccIdU + ".cif")
            #
            refTitle = refId + "/" + fitId
            fitTitle = fitId + "/" + refId
            self.__pairTupleList.append((refId, refPath, refTitle, fitId, fitPath, fitTitle))

    def __setPairIdList(self, pairIdList, cachePath="/data/components/ligand-dict-v3"):
        """Set the list of ID pais to be aligned by MCSS.

        From the input ID list build the internal pair list of
        tuples  [(refId,refPath,refTitle,fitId,fitPath,fitTitle),(),...]
        """

        self.__pairTupleList = []
        for refId, fitId in pairIdList:
            ccIdU = refId.upper()
            refPath = os.path.join(cachePath, ccIdU[0], ccIdU, ccIdU + ".cif")
            #
            ccIdU = fitId.upper()
            fitPath = os.path.join(cachePath, ccIdU[0], ccIdU, ccIdU + ".cif")
            #
            refTitle = refId + "/" + fitId
            fitTitle = fitId + "/" + refId
            self.__pairTupleList.append((refId, refPath, refTitle, fitId, fitPath, fitTitle))

    def getCCDefFile(self, ccFilePath, molBuildType="model-xyz", suppressHydrogens=False):
        """Fetch the molecule definition (ccPath) and build OE molecules
        for comparison.

        """
        #
        mU = MarshalUtil()
        rdCcObjL = mU.doImport(ccFilePath, fmt="mmcif")
        oemf = OeMoleculeFactory()
        ccId = oemf.setChemCompDef(rdCcObjL[0])
        oemf.build(molBuildType=molBuildType)

        if self.__verbose:
            logger.info("  CCId               = %s", ccId)
            logger.info("  Title              = %s", oemf.getTitle())
            logger.info("  SMILES             = %s", oemf.getCanSMILES())
            logger.info("  SMILES (stereo)    = %s", oemf.getIsoSMILES())
            logger.info("  Formula (Hill)     = %s", oemf.getFormula())
            logger.info("  InChI key          = %s", oemf.getInChIKey())
            logger.info("  InChI              = %s", oemf.getInChI())

        fD = {}
        fD = {"Formula": oemf.getFormula(), "SMILES": oemf.getCanSMILES(), "SMILES_STEREO": oemf.getIsoSMILES(), "InChI": oemf.getInChI(), "InChIKey": oemf.getInChIKey()}

        if suppressHydrogens:
            tMol = oemf.getGraphMolSuppressH()
        else:
            tMol = oemf.getMol()

        fD["OEMOL"] = tMol
        fD["xyz"] = oemf.getAtomDetails(xyzType="model")

        return (ccId, tMol, fD)

    def __getMiscFile(self, filePath, suppressHydrogens=False, importType="2D", title=None):
        """Fetch a miscellaneous chemical file (ccPath) and build OE molecules
        for comparison.

        """
        try:
            oeioU = OeIoUtils()
            oeMolL = oeioU.fileToMols(filePath, use3D=importType == "3D")
            oeMol = oeMolL[0]

            ccId = title if title else oeMol.GetTitle()
            if title:
                oeMol.SetTitle(ccId)
            #
            oemf = OeMoleculeFactory()
            oemf.setOeMol(oeMol, ccId)
            #
            fD = oemf.getOeMoleculeFeatures()
            if self.__verbose:
                logger.info("  Title              = %s", title)
                logger.info("  Title OEMF         = %s", oemf.getTitle())
                logger.info("  SMILES             = %s", oemf.getCanSMILES())
                logger.info("  SMILES (stereo)    = %s", oemf.getIsoSMILES())
                logger.info("  Formula (Hill)     = %s", oemf.getFormula())
                logger.info("  InChI key          = %s", oemf.getInChIKey())
                logger.info("  InChI              = %s", oemf.getInChI())
            #
            ccId = oemf.getTitle()
            if suppressHydrogens:
                tMol = oemf.getGraphMolSuppressH()
            else:
                tMol = oemf.getMol()

            molXyzL = []
            if importType == "3D":
                for atm in tMol.GetAtoms():
                    xyzL = oechem.OEFloatArray(3)
                    tMol.GetCoords(atm, xyzL)
                    molXyzL.append(
                        ComponentAtomDetails(
                            atIdx=atm.GetIdx(),
                            atNo=atm.GetAtomicNum(),
                            atName=atm.GetName(),
                            atType=atm.GetType(),
                            x=xyzL[0],
                            y=xyzL[1],
                            z=xyzL[2],
                            atFormalCharge=atm.GetFormalCharge(),
                        )
                    )
            fD = {}
            fD = {
                "Formula": oemf.getFormula(),
                "SMILES": oemf.getCanSMILES(),
                "SMILES_STEREO": oemf.getIsoSMILES(),
                "InChI": oemf.getInChI(),
                "InChIKey": oemf.getInChIKey(),
                "xyz": molXyzL,
            }
            for atm in tMol.GetAtoms():
                xyzL = oechem.OEFloatArray(3)
                tMol.GetCoords(atm, xyzL)
                if self.__verbose:
                    logger.debug("atom  %s %s %s %s %r", atm.GetIdx(), atm.GetAtomicNum(), atm.GetName(), atm.GetType(), xyzL)

            fD["OEMOL"] = tMol
            return (ccId, tMol, fD)
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return None, None, None

    def __setupMCSS(self, refmol, useExhaustive=True):
        """Internal initialization for the MCSS comparison."""
        #
        mode = oechem.OEMCSType_Exhaustive if useExhaustive else oechem.OEMCSType_Approximate
        self.__mcss = oechem.OEMCSSearch(mode)
        atomexpr, bondexpr = OeCommonUtils.getAtomBondExprOpts(self.__searchType)
        self.__mcss.Init(refmol, atomexpr, bondexpr)
        if self.__verbose:
            logger.info("Initialize MCSS (%r)", self.__searchType)

    def __setupSubStructure(self, refmol):
        """Internal initialization for a substructure comparison."""
        #
        atomexpr, bondexpr = OeCommonUtils.getAtomBondExprOpts(self.__searchType)
        self.__ss = oechem.OESubSearch(refmol, atomexpr, bondexpr)
        if self.__verbose:
            logger.info("Initialize SS (%r)", self.__searchType)

    @timeout(100)
    def doAlignSs(self, unique=True):
        """Test the SS comparison between current reference and fit molecules -
        Return list of corresponding atoms on success or an empty list otherwise.
        """
        atomMapL = []
        fitAtomUnMappedL = []
        #
        nAtomsRef = self.__refmol.NumAtoms()
        nAtomsFit = self.__fitmol.NumAtoms()
        # -------
        oechem.OEAddExplicitHydrogens(self.__refmol)
        oechem.OEAddExplicitHydrogens(self.__fitmol)
        fitAtD = {}
        #
        for at in self.__fitmol.GetAtoms():
            nAtL = at.GetAtoms()
            neighbors = [nAt.GetName() for nAt in nAtL]
            atType = oechem.OEGetAtomicSymbol(at.GetAtomicNum())
            xyzL = oechem.OEFloatArray(3)
            self.__fitmol.GetCoords(at, xyzL)
            fitAtD[at.GetIdx()] = AlignAtomUnMapped(
                fitId=self.__fitId,
                fitAtIdx=at.GetIdx(),
                fitAtName=at.GetName(),
                fitAtType=atType,
                fitAtNo=at.GetAtomicNum(),
                fitAtFormalCharge=at.GetFormalCharge(),
                x=xyzL[0],
                y=xyzL[1],
                z=xyzL[2],
                fitNeighbors=neighbors,
            )

        #
        logger.debug("nAtomsRef %d nAtomsFit %d", nAtomsRef, nAtomsFit)
        #
        # --------
        self.__setupSubStructure(self.__refmol)
        miter = self.__ss.Match(self.__fitmol, unique)
        if miter.IsValid():
            match = miter.Target()
            for mAt in match.GetAtoms():
                atomMapL.append(
                    AlignAtomMap(
                        refId=self.__refId,
                        refAtIdx=mAt.pattern.GetIdx(),
                        refAtNo=mAt.pattern.GetAtomicNum(),
                        refAtName=mAt.pattern.GetName(),
                        fitId=self.__fitId,
                        fitAtIdx=mAt.target.GetIdx(),
                        fitAtNo=mAt.target.GetAtomicNum(),
                        fitAtName=mAt.target.GetName(),
                    )
                )
                fitAtD.pop(mAt.target.GetIdx())
            logger.debug("fitAtD %r", fitAtD)
        fitAtomUnMappedL = list(fitAtD.values())
        return (nAtomsRef, self.__refFD, nAtomsFit, self.__fitFD, atomMapL, fitAtomUnMappedL)

    @timeout(100)
    def doAlignMcss(self, unique=True, minFrac=1.0, useExhaustive=True):
        """Test the MCSS comparison between current reference and fit molecules -
        Return list of corresponding atoms on success or an empty list otherwise.
        """
        atomMapL = []
        fitAtomUnMappedL = []
        #
        nAtomsRef = self.__refmol.NumAtoms()
        nAtomsFit = self.__fitmol.NumAtoms()
        fitAtD = {}
        for at in self.__fitmol.GetAtoms():
            nAtL = at.GetAtoms()
            neighbors = [nAt.GetName() for nAt in nAtL]
            atType = oechem.OEGetAtomicSymbol(at.GetAtomicNum())
            xyzL = oechem.OEFloatArray(3)
            self.__fitmol.GetCoords(at, xyzL)
            fitAtD[at.GetIdx()] = AlignAtomUnMapped(
                fitId=self.__fitId,
                fitAtIdx=at.GetIdx(),
                fitAtName=at.GetName(),
                fitAtType=atType,
                fitAtNo=at.GetAtomicNum(),
                fitAtFormalCharge=at.GetFormalCharge(),
                x=xyzL[0],
                y=xyzL[1],
                z=xyzL[2],
                fitNeighbors=neighbors,
            )
        minAtoms = int(min(nAtomsRef, nAtomsFit) * minFrac)
        # -------
        self.__setupMCSS(self.__refmol, useExhaustive=useExhaustive)
        self.__mcss.SetMCSFunc(oechem.OEMCSMaxAtoms())
        self.__mcss.SetMinAtoms(minAtoms)
        oechem.OEAddExplicitHydrogens(self.__refmol)
        oechem.OEAddExplicitHydrogens(self.__fitmol)
        #
        # --------
        miter = self.__mcss.Match(self.__fitmol, unique)
        if miter.IsValid():
            match = miter.Target()
            for mAt in match.GetAtoms():

                atomMapL.append(
                    AlignAtomMap(
                        refId=self.__refId,
                        refAtIdx=mAt.pattern.GetIdx(),
                        refAtNo=mAt.pattern.GetAtomicNum(),
                        refAtName=mAt.pattern.GetName(),
                        fitId=self.__fitId,
                        fitAtIdx=mAt.target.GetIdx(),
                        fitAtNo=mAt.target.GetAtomicNum(),
                        fitAtName=mAt.target.GetName(),
                    )
                )
                fitAtD.pop(mAt.target.GetIdx())
        fitAtomUnMappedL = list(fitAtD.values())
        return (nAtomsRef, self.__refFD, nAtomsFit, self.__fitFD, atomMapL, fitAtomUnMappedL)



##
# File:    RdMoleculeFactory.py
# Author:  jdw
# Date:    1-Jan-2020
# Version: 0.001
#
# Updates:
#
##
# flake8: noqa
# pylint: skip-file
"""
Classes to build RdKit molecule objects from chemical component definition data.
"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import logging
from collections import defaultdict, namedtuple

from rdkit import Chem
from rdkit.Chem.rdMolDescriptors import CalcMolFormula

logger = logging.getLogger(__name__)

ComponentDetails = namedtuple("ComponentDetails", "ccId formula ifCharge")
ComponentAtom = namedtuple("ComponentAtom", "name aType isAromatic isChiral CIP fCharge")
ComponentBond = namedtuple("ComponentBond", "iType isAromatic CIP")
ComponentDescriptors = namedtuple("ComponentDescriptors", "smiles isoSmiles inchi inchiKey")


class RdMoleculeFactory(object):
    """Utility methods for constructing RdKit molecules from chemical component definition objects."""

    def __init__(self, verbose=False):
        self.__verbose = verbose
        self.__ccId = None
        self.__rdMol = None
        self.__atomIdxD = None

    def setFile(self, ccId, filePath, atomIdxD=None, molFormat="sdf"):
        try:
            self.__ccId = ccId
            self.__atomIdxD = atomIdxD if atomIdxD else {}
            if molFormat in ["sdf", "mol"]:
                self.__rdMol = Chem.MolFromMolFile(filePath)
                return True
            else:
                logger.error("Unsupported format %r", molFormat)
        except Exception as e:
            logger.exception("Failing for %s with %s", ccId, str(e))
        return False

    def updateProperties(self):
        if not self.__rdMol:
            return False
        try:
            self.__rdMol.UpdatePropertyCache()
            for atom in self.__rdMol.GetAtoms():
                atom.UpdatePropertyCache()
            return True
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return False

    def updateStereoChemAssignments(self):
        if not self.__rdMol:
            return False
        try:
            Chem.AssignAtomChiralTagsFromStructure(self.__rdMol)
            Chem.AssignStereochemistry(self.__rdMol, True, True, True)
            # self.__rdMol.Debug()
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return False

    def updateAromaticAssignments(self, aromaticModel="default"):
        mD = {
            "simple": Chem.AromaticityModel.AROMATICITY_SIMPLE,
            "default": Chem.AromaticityModel.AROMATICITY_DEFAULT,
            "mdl": Chem.AromaticityModel.AROMATICITY_MDL,
            "rdkit": Chem.AromaticityModel.AROMATICITY_RDKIT,
        }
        if not self.__rdMol or aromaticModel not in mD:
            return False
        try:
            #  self.__rdMol.Debug()
            Chem.SetAromaticity(self.__rdMol, mD[aromaticModel])
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return True

    def getMoleculeFeatures(self):
        """Get the essential features of the constructed rdMol for the input component."""
        mD = self.__rdMol.GetPropsAsDict()
        logger.debug("mol props %r", mD.items())
        #
        formula = CalcMolFormula(self.__rdMol)
        ccId = self.__ccId
        ifCharge = Chem.rdmolops.GetFormalCharge(self.__rdMol)
        #
        inchiKey = Chem.inchi.MolToInchiKey(self.__rdMol)
        inchi = Chem.inchi.MolToInchi(self.__rdMol)
        smiles = Chem.rdmolfiles.MolToSmiles(self.__rdMol, isomericSmiles=False, canonical=True)
        isoSmiles = Chem.rdmolfiles.MolToSmiles(self.__rdMol, isomericSmiles=True, canonical=True)
        logger.debug("%s formula %s", ccId, formula)
        details = ComponentDetails(ccId=ccId, formula=formula, ifCharge=ifCharge)
        descriptors = ComponentDescriptors(smiles=smiles, isoSmiles=isoSmiles, inchi=inchi, inchiKey=inchiKey)
        #
        typeCounts = defaultdict(int)
        ccAtomD = {}
        ccAtomIdD = {}
        for ii, at in enumerate(self.__rdMol.GetAtoms(), 1):
            atIdx = at.GetIdx()
            aType = at.GetSymbol()
            typeCounts[aType] += 1
            atName = self.__atomIdxD[ii] if ii in self.__atomIdxD else aType + str(typeCounts[aType])
            # atNo = at.GetAtomicNum()
            isAromatic = at.GetIsAromatic()
            isChiral = at.GetChiralTag() > 0
            iCharge = at.GetFormalCharge()
            # cipStereo = at.GetProp("_CIPCode")
            atD = at.GetPropsAsDict(includePrivate=True, includeComputed=True)
            cipStereo = None
            if "_CIPCode" in atD:
                cipStereo = atD["_CIPCode"]
            if cipStereo and cipStereo not in ["S", "R"]:
                logger.error("%s (%s): Unexpected atom CIP stereo setting %r", ccId, atName, cipStereo)
            #
            ccAtomD[atName] = ComponentAtom(name=atName, aType=aType, isAromatic=isAromatic, isChiral=isChiral, CIP=cipStereo, fCharge=iCharge)
            ccAtomIdD[atIdx] = atName
            # nL = at.GetProp(includePrivate=True, includeComputed=True)
            atD = at.GetPropsAsDict(includePrivate=True, includeComputed=True)
            logger.debug("%s Atom %s %s %r %r %s", ccId, atName, aType, isAromatic, isChiral, cipStereo)
        #
        ccBondD = {}
        for bnd in self.__rdMol.GetBonds():
            atI = bnd.GetBeginAtomIdx()
            atJ = bnd.GetEndAtomIdx()
            atNameI = ccAtomIdD[atI]
            atNameJ = ccAtomIdD[atJ]
            isAromatic = bnd.GetIsAromatic()
            #
            # bType = bnd.GetBondType()
            # iType = 0
            cipStereo = None
            tS = bnd.GetStereo()
            if tS == Chem.rdchem.BondStereo.STEREOE:
                cipStereo = "E"
            elif tS == Chem.rdchem.BondStereo.STEREOZ:
                cipStereo = "Z"

            # bL = bnd.GetPropNames(includePrivate=True, includeComputed=True)
            bD = bnd.GetPropsAsDict(includePrivate=True, includeComputed=True)
            iType = bD["_MolFileBondType"]
            logger.debug("Bond %s %s iType %r cipStereo %r aromatic %r", atNameI, atNameJ, iType, cipStereo, isAromatic)
            #
            if cipStereo and cipStereo not in ["E", "Z"]:
                logger.error("%s (%s %s): Unexpected bond CIP stereo setting %r", ccId, atNameI, atNameJ, cipStereo)
            #
            ccBondD[(atNameI, atNameJ)] = ComponentBond(iType=iType, isAromatic=isAromatic, CIP=cipStereo)
        #
        ccD = {"details": details, "descriptors": descriptors, "atoms": ccAtomD, "bonds": ccBondD}
        return ccD



##
# File: MolecularFormula.py
# Date: 14-Mar-2020
#
# Adapted from recipe a recipe posted by at:
# https://github.com/Zapaan/python-chemical-formula-parser
##

import logging
import re
from collections import Counter

logger = logging.getLogger(__name__)


class MolecularFormula(object):
    def __init__(self):
        self.__regex = r"([A-Z][a-z]*)(\d*)"
        self.__openGroup = "({["
        self.__closeGroup = ")}]"

    def __isBalanced(self, formula):
        """Check if all grouping elements are properly paired by count."""
        cnt = Counter(formula)
        return cnt["["] == cnt["]"] and cnt["{"] == cnt["}"] and cnt["("] == cnt[")"]

    def __toDict(self, tuples):
        """Transform tuples of tuples to a dict of elements."""
        res = dict()
        for atom, num in tuples:
            try:
                res[atom] += int(num or 1)
            except KeyError:
                res[atom] = int(num or 1)
        return res

    def __mergeDict(self, mol1, mol2, weight=1):
        """
        Merge 2 dicts representing molecules. Return a new dict.
        """
        return {atom: (mol1.get(atom, 0) + mol2.get(atom, 0)) * weight for atom in set(mol1) | set(mol2)}

    def __parse(self, formula):
        """
        Return an element dictionary and length of parsed part.

        Recurse on grouped elements to parse the subpart.
        """
        qL = []
        mol = {}
        ii = 0

        while ii < len(formula):
            # Using a classic loop allow for manipulating the cursor
            token = formula[ii]

            if token in self.__closeGroup:
                # Check for an index for this part
                mtch = re.match(r"\d+", formula[ii + 1 :])
                if mtch:
                    weight = int(mtch.group(0))
                    ii += len(mtch.group(0))
                else:
                    weight = 1

                submol = self.__toDict(re.findall(self.__regex, "".join(qL)))
                return self.__mergeDict(mol, submol, weight), ii

            elif token in self.__openGroup:
                submol, ll = self.__parse(formula[ii + 1 :])
                mol = self.__mergeDict(mol, submol)
                # skip the already read submol
                ii += ll + 1
            else:
                qL.append(token)

            ii += 1
        # merge in all that's left at base level
        return self.__mergeDict(mol, self.__toDict(re.findall(self.__regex, "".join(qL)))), ii

    def parseFormula(self, formula):
        """Parse the formula and return a dictionary of counts of each atom type."""
        rD = self.__parse(formula)[0]
        logger.debug("formula %r result %r", formula, rD)
        return rD



##
# File:    ObMoleculeFactory.py
# Author:  jdw
# Date:    11-Jan-2020
# Version: 0.001
#
# Updates:
#
##
# flake8: noqa
# pylint: skip-file
"""
Classes to build OpenBabel molecule objects from chemical component definition data.
"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import logging
from collections import defaultdict, namedtuple

from openbabel import openbabel
from openbabel import pybel


logger = logging.getLogger(__name__)

ComponentDetails = namedtuple("ComponentDetails", "ccId formula ifCharge")
ComponentAtom = namedtuple("ComponentAtom", "name aType isAromatic isChiral CIP fCharge")
ComponentBond = namedtuple("ComponentBond", "iType isAromatic CIP")
ComponentDescriptors = namedtuple("ComponentDescriptors", "smiles isoSmiles inchi inchiKey")


class ObMoleculeFactory(object):
    """Utility methods for constructing RdKit molecules from chemical component definition objects."""

    def __init__(self, verbose=False):
        self.__verbose = verbose
        self.__ccId = None
        self.__pybelMol = None
        self.__atomIdxD = None
        #
        self.__obConv = openbabel.OBConversion()
        self.__inputFormatDict = dict([f.split(" -- ") for f in self.__obConv.GetSupportedInputFormat()])
        self.__outputFormatDict = dict([f.split(" -- ") for f in self.__obConv.GetSupportedOutputFormat()])

    def setFile(self, ccId, filePath, molFormat="mol", atomIdxD=None):
        try:
            if molFormat not in self.__inputFormatDict:
                return False
            self.__ccId = ccId
            self.__atomIdxD = atomIdxD
            self.__pybelMol = next(pybel.readfile(molFormat, filePath), None)
            logger.info("pybelMol type %r", self.__pybelMol)
            return True
        except Exception as e:
            logger.exception("Failing for %s with %s", ccId, str(e))
        return False

    def setString(self, ccId, molText, molFormat="mol", atomIdxD=None):
        try:
            if molFormat not in self.__inputFormatDict:
                return False
            self.__ccId = ccId
            self.__atomIdxD = atomIdxD
            self.__pybelMol = pybel.readstring(molFormat, molText)
            logger.info("obMol type %r", self.__pybelMol)
            return True
        except Exception as e:
            logger.exception("Failing for %s with %s", ccId, str(e))
        return False

    def getMoleculeFeatures(self):
        """Get the essential features of the constructed obMol for the input component.

        OBConversion object, use SetInAndOutFormat(InCode, OutCode).
        To set a Read Option s, use SetOptions("s", OBConversion::INOPTIONS).

        """
        title = self.__pybelMol.title
        molWeight = self.__pybelMol.molwt
        formula = self.__pybelMol.formula
        ccId = title
        ifCharge = self.__pybelMol.charge
        logger.info("%s formula %s charge %d mw %f", title, formula, ifCharge, molWeight)
        inchi = self.__pybelMol.write("inchi").strip()
        inchiKey = self.__pybelMol.write("inchikey").strip()
        smiles = self.__pybelMol.write("can", opt={"n": None}).strip()
        isoSmiles = self.__pybelMol.write("can", opt={"i": None, "n": None}).strip()
        details = ComponentDetails(ccId=ccId, formula=formula, ifCharge=ifCharge)
        descriptors = ComponentDescriptors(smiles=smiles, isoSmiles=isoSmiles, inchi=inchi, inchiKey=inchiKey)
        #
        #
        typeCounts = defaultdict(int)
        ccAtomD = {}
        ccAtomIdD = {}
        for ii, pat in enumerate(self.__pybelMol.atoms, 1):
            at = pat.OBAtom
            atIdx = at.GetIdx()
            # atNo = at.GetAtomicNum()
            aType = at.GetType()
            typeCounts[aType] += 1
            atName = self.__atomIdxD[ii] if ii in self.__atomIdxD else aType + str(typeCounts[aType])
            #
            isAromatic = at.IsAromatic()
            isChiral = at.IsChiral()
            iCharge = at.GetFormalCharge()
            cipStereo = None
            ccAtomD[atName] = ComponentAtom(name=atName, aType=aType, isAromatic=isAromatic, isChiral=isChiral, CIP=cipStereo, fCharge=iCharge)
            ccAtomIdD[atIdx] = atName
            logger.debug("%s Atom %s %s %r %r %s", ccId, atName, aType, isAromatic, isChiral, cipStereo)
        #
        ccBondD = {}
        for bnd in openbabel.OBMolBondIter(self.__pybelMol.OBMol):
            atI = bnd.GetBeginAtomIdx()
            atJ = bnd.GetEndAtomIdx()
            atNameI = ccAtomIdD[atI]
            atNameJ = ccAtomIdD[atJ]
            isAromatic = bnd.IsAromatic()
            iType = bnd.GetBondOrder()
            cipStereo = None
            logger.debug("Bond %s %s iType %r cipStereo %r aromatic %r", atNameI, atNameJ, iType, cipStereo, isAromatic)
            #
            ccBondD[(atNameI, atNameJ)] = ComponentBond(iType=iType, isAromatic=isAromatic, CIP=cipStereo)
        #
        ccD = {"details": details, "descriptors": descriptors, "atoms": ccAtomD, "bonds": ccBondD}
        return ccD



##
# File:    OeChemCompUtils.py
# Author:  jdw
# Date:    17-Jan-2021
# Version: 0.001
#
# Updates:
#
#
##
"""
Utilities to build chemical component definitions from OE molecule objects.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import logging
import time

from openeye import oechem
from openeye import oeiupac

from mmcif.api.DataCategory import DataCategory
from mmcif.api.PdbxContainers import DataContainer
from rcsb.utils.io.MarshalUtil import MarshalUtil


logger = logging.getLogger(__name__)


class OeChemCompUtils(object):
    """Build chemical component definitions from OE molecule objects."""

    def __init__(self):

        self.__oeVersion = oechem.OEToolkitsGetRelease()
        self.__containerList = []

    def write(self, filePath):
        """Write the contents of the current data container of chemical component definitions.

        Args:
            filePath (str): output chemical component definition file path

        Returns:
            (bool): True for success or False otherwise
        """
        try:
            mU = MarshalUtil()
            ok = mU.doExport(filePath, self.__containerList, fmt="mmcif")
        except Exception as e:
            logger.exception("Failing %r with %s", filePath, str(e))
            ok = False
        return ok

    def addOeMol(self, ccId, oeMol, missingModelXyz=True, writeIdealXyz=False):
        """Add the input oeMol to the current PDBx data container as a chemical component definition.

        Args:
            oeMol (obj): instance of OE molecule
            ccId (str): chemical component identifer
            name (str, optional): chemical component name. Defaults to None.
            missingModelXyz (bool, optional): set the missing model coordinate flag. Defaults to True.
            writeIdealXyz (bool, optional): write 3D coordinates in using ideal coordinate data items. Defaults to False.

        Returns:
            (bool): True for success and False otherwise
        """
        try:
            ccIdU = str(ccId).strip().upper()
            curContainer = DataContainer(ccIdU)
            #
            rowD = self.__makeChemCompCategory(ccIdU, oeMol, site="RCSB", missingModelXyz=missingModelXyz)
            aCat = DataCategory("chem_comp", list(rowD.keys()), [rowD])
            curContainer.append(aCat)
            #
            rowDL = self.__makeChemCompAtomCategory(ccIdU, oeMol, writeIdealXyz=writeIdealXyz)
            aCat = DataCategory("chem_comp_atom", list(rowDL[0].keys()), rowDL)
            curContainer.append(aCat)
            #
            rowDL = self.__makeChemCompBondCategory(ccIdU, oeMol)
            aCat = DataCategory("chem_comp_bond", list(rowDL[0].keys()), rowDL)
            curContainer.append(aCat)
            #
            rowDL = self.__makeChemCompDescriptorCategory(ccIdU, oeMol)
            aCat = DataCategory("pdbx_chem_comp_descriptor", list(rowDL[0].keys()), rowDL)
            curContainer.append(aCat)
            #
            rowDL = self.__makeChemCompIdentifierCategory(ccIdU, oeMol)
            aCat = DataCategory("pdbx_chem_comp_identifier", list(rowDL[0].keys()), rowDL)
            curContainer.append(aCat)
            #
            rowD = self.__makeChemCompAuditRow(ccIdU)
            aCat = DataCategory("pdbx_chem_comp_audit", list(rowD.keys()), [rowD])
            curContainer.append(aCat)
            #
            self.__containerList.append(curContainer)
            return True
        except Exception as e:
            logger.exception("Failing %r with %s", ccId, str(e))
            #
        return False

    def __getFormalCharge(self, oeMol):
        fCharge = 0
        for atom in oeMol.GetAtoms():
            fCharge += atom.GetFormalCharge()
        return fCharge

    def __makeChemCompAuditRow(self, ccId, action="CREATE", date=None, processingSite="RCSB", annotator="?", details="?"):
        """
        loop_
        _pdbx_chem_comp_audit.comp_id
        _pdbx_chem_comp_audit.action_type
        _pdbx_chem_comp_audit.date
        _pdbx_chem_comp_audit.processing_site
        _pdbx_chem_comp_audit.annotator
        _pdbx_chem_comp_audit.details
        ARG "Create component"  1999-07-08 RCSB ? ?

        """
        aRow = {}
        aRow["comp_id"] = ccId
        aRow["action_type"] = action
        if date is None:
            date = time.strftime("%Y-%m-%d", time.localtime())
        aRow["date"] = date
        aRow["processing_site"] = processingSite
        aRow["annotator"] = annotator
        aRow["details"] = details
        #
        return aRow

    def __makeChemCompDescriptorCategory(self, ccId, oeMol):
        """
            loop_
            _pdbx_chem_comp_descriptor.comp_id
            _pdbx_chem_comp_descriptor.type
            _pdbx_chem_comp_descriptor.program
            _pdbx_chem_comp_descriptor.program_version
            _pdbx_chem_comp_descriptor.descriptor
            ARG SMILES           ACDLabs              10.04 "O=C(O)C(N)CCCNC(=[NH2+])N"
            ARG SMILES_CANONICAL CACTVS               3.341 "N[C@@H](CCCNC(N)=[NH2+])C(O)=O"
            ARG SMILES           CACTVS               3.341 "N[CH](CCCNC(N)=[NH2+])C(O)=O"
            ARG SMILES_CANONICAL "OpenEye OEToolkits" 1.5.0 "C(C[C@@H](C(=O)O)N)CNC(=[NH2+])N"
            ARG SMILES           "OpenEye OEToolkits" 1.5.0 "C(CC(C(=O)O)N)CNC(=[NH2+])N"
            ARG InChI            InChI                1.03  "InChI=1S/C6H14N4O2/c7-4(5(11)12)2-1-3-1..... "
            ARG InChIKey         InChI                1.03  ODKSFYDXXFIFQN-BYPYZUCNSA-O
        #
        """
        rowL = []
        #
        aRow = {}
        aRow["comp_id"] = ccId
        aRow["type"] = "SMILES_CANONICAL"
        aRow["program"] = "OpenEye OEToolkits"
        aRow["program_version"] = self.__oeVersion
        aRow["descriptor"] = oechem.OECreateIsoSmiString(oeMol)
        rowL.append(aRow)
        #
        aRow = {}
        aRow["comp_id"] = ccId
        aRow["type"] = "SMILES"
        aRow["program"] = "OpenEye OEToolkits"
        aRow["program_version"] = self.__oeVersion
        aRow["descriptor"] = oechem.OECreateCanSmiString(oeMol)
        rowL.append(aRow)
        #
        aRow = {}
        aRow["comp_id"] = ccId
        aRow["type"] = "InChI"
        aRow["program"] = "OpenEye OEToolkits"
        aRow["program_version"] = self.__oeVersion
        aRow["descriptor"] = oechem.OECreateInChI(oeMol)
        rowL.append(aRow)
        #
        aRow = {}
        aRow["comp_id"] = ccId
        aRow["type"] = "InChIKey"
        aRow["program"] = "OpenEye OEToolkits"
        aRow["program_version"] = self.__oeVersion
        aRow["descriptor"] = oechem.OECreateInChIKey(oeMol)
        rowL.append(aRow)
        #
        return rowL

    def __makeChemCompIdentifierCategory(self, ccId, oeMol):
        """

        loop_
        _pdbx_chem_comp_identifier.comp_id
        _pdbx_chem_comp_identifier.type
        _pdbx_chem_comp_identifier.program
        _pdbx_chem_comp_identifier.program_version
        _pdbx_chem_comp_identifier.identifier
        ATP "SYSTEMATIC NAME" ACDLabs              10.04
        ;adenosine 5'-(tetrahydrogen triphosphate)
        ;
        ATP "SYSTEMATIC NAME" "OpenEye OEToolkits" 1.5.0 "[[(2R,3S,4R,5R)-5-(6-aminopurin-9-yl)-..."
        #
        """
        rowL = []
        #
        aRow = {}
        aRow["comp_id"] = ccId
        aRow["type"] = "SYSTEMATIC NAME"
        aRow["program"] = "OpenEye OEToolkits"
        aRow["program_version"] = self.__oeVersion
        style = oeiupac.OEGetIUPACNamStyle("systematic")
        name = oeiupac.OEToUTF8(oeiupac.OECreateIUPACName(oeMol, style))
        aRow["identifier"] = name
        rowL.append(aRow)
        aRow = {}
        aRow["comp_id"] = ccId
        aRow["type"] = "COMMON"
        aRow["program"] = "OpenEye OEToolkits"
        aRow["program_version"] = self.__oeVersion
        style = oeiupac.OEGetIUPACNamStyle("traditional")
        name = oeiupac.OEToUTF8(oeiupac.OECreateIUPACName(oeMol, style))
        aRow["identifier"] = name
        rowL.append(aRow)
        #
        aRow = {}
        aRow["comp_id"] = ccId
        aRow["type"] = "SYNONYM"
        aRow["program"] = "OpenEye OEToolkits"
        aRow["program_version"] = self.__oeVersion
        style = oeiupac.OEGetIUPACNamStyle("acdname")
        name = oeiupac.OEToUTF8(oeiupac.OECreateIUPACName(oeMol, style))
        aRow["identifier"] = name
        rowL.append(aRow)
        #
        return rowL

    def __makeChemCompCategory(self, ccId, oeMol, site="RCSB", missingModelXyz=False):
        #
        lt = time.strftime("%Y-%m-%d", time.localtime())
        formula = oechem.OEMolecularFormula(oeMol)
        charge = self.__getFormalCharge(oeMol)
        fW = oechem.OECalculateMolecularWeight(oeMol)
        style = oeiupac.OEGetIUPACNamStyle("systematic")
        name = oeiupac.OEToUTF8(oeiupac.OECreateIUPACName(oeMol, style))
        #
        ccRow = {}
        ccRow["id"] = ccId
        if name is not None:
            ccRow["name"] = name
        else:
            ccRow["name"] = "?"
        ccRow["type"] = "NON-POLYMER"
        ccRow["pdbx_type"] = "?"
        if formula is not None:
            ccRow["formula"] = formula
        else:
            ccRow["formula"] = "?"
        ccRow["mon_nstd_parent_comp_id"] = "?"
        # ccRow["pdbx_synonyms"] = "?"
        if charge is not None:
            ccRow["pdbx_formal_charge"] = charge
        else:
            ccRow["pdbx_formal_charge"] = "?"
        ccRow["pdbx_ambiguous_flag"] = "N"
        ccRow["pdbx_initial_date"] = lt
        ccRow["pdbx_modified_date"] = lt
        ccRow["pdbx_release_status"] = "HOLD"
        ccRow["pdbx_replaced_by"] = "?"
        ccRow["pdbx_replaces"] = "?"
        if fW is not None:
            ccRow["formula_weight"] = "%0.3f" % fW
        else:
            ccRow["formula_weight"] = "?"
        ccRow["one_letter_code"] = "?"
        tlc = ccId.split("_")[0]
        ccRow["three_letter_code"] = tlc
        ccRow["pdbx_model_coordinates_details"] = "?"
        ccRow["pdbx_ideal_coordinates_details"] = "?"
        if missingModelXyz:
            ccRow["pdbx_model_coordinates_missing_flag"] = "Y"
        else:
            ccRow["pdbx_model_coordinates_missing_flag"] = "N"
        ccRow["pdbx_model_coordinates_db_code"] = "?"
        ccRow["pdbx_processing_site"] = site
        ccRow["pdbx_subcomponent_list"] = "?"
        return ccRow

    def __makeChemCompAtomCategory(self, ccId, oeMol, writeIdealXyz=True):
        """Populate elements of chemical component definition for atoms and bonds."""
        #
        idCode = ccId
        #
        rowL = []
        for ii, atom in enumerate(oeMol.GetAtoms()):
            #
            atRow = {}
            atRow["comp_id"] = idCode
            atRow["atom_id"] = atom.GetName().strip()
            atRow["alt_atom_id"] = atom.GetName().strip()
            atRow["type_symbol"] = oechem.OEGetAtomicSymbol(atom.GetAtomicNum())
            atRow["charge"] = atom.GetFormalCharge()
            (xC, yC, zC) = oeMol.GetCoords(atom)
            if writeIdealXyz:
                atRow["pdbx_model_Cartn_x_ideal"] = "%0.3f" % xC
                atRow["pdbx_model_Cartn_y_ideal"] = "%0.3f" % yC
                atRow["pdbx_model_Cartn_z_ideal"] = "%0.3f" % zC
            else:
                atRow["model_Cartn_x"] = "%0.3f" % xC
                atRow["model_Cartn_y"] = "%0.3f" % yC
                atRow["model_Cartn_z"] = "%0.3f" % zC
            if atom.GetStringData("pdbx_leaving_atom_flag") in ["Y", "N"]:
                atRow["pdbx_leaving_atom_flag"] = atom.GetStringData("pdbx_leaving_atom_flag")
            else:
                atRow["pdbx_leaving_atom_flag"] = "N"
            if len(atRow["atom_id"]) > 3 or len(atRow["type_symbol"]) == 2:
                atRow["pdbx_align"] = 0
            else:
                atRow["pdbx_align"] = 1

            if atom.IsAromatic():
                atRow["pdbx_aromatic_flag"] = "Y"
            else:
                atRow["pdbx_aromatic_flag"] = "N"
            # oeSt = oechem.OEGetCIPStereo(mol, atom)
            oeSt = None
            # if atom.IsChiral():
            cip = oechem.OEPerceiveCIPStereo(oeMol, atom)
            if atom.HasStereoSpecified():
                if cip == oechem.OECIPAtomStereo_S:
                    oeSt = "S"
                if cip == oechem.OECIPAtomStereo_R:
                    oeSt = "R"
                if cip == oechem.OECIPAtomStereo_NotStereo:
                    oeSt = None
                if cip == oechem.OECIPAtomStereo_UnspecStereo:
                    oeSt = "U"
                # oeSt = oechem.OEGetCIPStereo(oeMol, atom)
                # oeSt = atom.GetCIPStereo()
            if oeSt is not None and (len(oeSt) > 0) and (oeSt == "R" or oeSt == "S"):
                atRow["pdbx_stereo_config"] = oeSt
            else:
                atRow["pdbx_stereo_config"] = "N"
            atRow["pdbx_component_atom_id"] = atom.GetName().strip()
            atRow["pdbx_component_comp_id"] = idCode
            atRow["pdbx_ordinal"] = str(ii + 1)
            rowL.append(atRow)
        #
        return rowL

    def __makeChemCompBondCategory(self, ccId, oeMol):
        idCode = ccId
        rowL = []
        iBondD = {1: "SING", 2: "DOUB", 3: "TRIP", 4: "QUAD", 5: "AROM", 0: "DELO"}

        for ii, bond in enumerate(oeMol.GetBonds()):
            oeOrder = str(bond.GetOrder()).upper()
            b1 = bond.GetBgn().GetName().strip()
            b2 = bond.GetEnd().GetName().strip()
            if (b1 is None) or (len(b1) < 1):
                continue
            if (b2 is None) or (len(b1) < 1):
                continue
            if int(oeOrder) in iBondD:
                bndRow = {}
                bndRow["comp_id"] = idCode
                bndRow["atom_id_1"] = b1
                bndRow["atom_id_2"] = b2
                bndRow["value_order"] = iBondD[int(oeOrder)]

                if bond.IsAromatic():
                    bndRow["pdbx_aromatic_flag"] = "Y"
                else:
                    bndRow["pdbx_aromatic_flag"] = "N"

                oeSt = None
                if bond.GetOrder() == 2:
                    cip = oechem.OEPerceiveCIPStereo(oeMol, bond)
                    if bond.HasStereoSpecified():
                        if cip == oechem.OECIPBondStereo_E:
                            oeSt = "E"
                        if cip == oechem.OECIPBondStereo_Z:
                            oeSt = "Z"
                        if cip == oechem.OECIPBondStereo_NotStereo:
                            oeSt = None
                        if cip == oechem.OECIPBondStereo_UnspecStereo:
                            oeSt = None
                if oeSt is not None and (len(oeSt) > 0) and (oeSt == "E" or oeSt == "Z"):
                    bndRow["pdbx_stereo_config"] = oeSt
                else:
                    bndRow["pdbx_stereo_config"] = "N"

                bndRow["pdbx_ordinal"] = str(ii + 1)
                rowL.append(bndRow)
        return rowL

    def __getElementCounts(self, oeMol):
        """Get the dictionary of element counts (eg. eD[iAtNo]=iCount)."""
        eD = {}
        if len(eD) == 0:
            # calculate from current oeMol
            try:
                eD = {}
                for atom in oeMol.GetAtoms():
                    atNo = atom.GetAtomicNum()
                    if atNo not in eD:
                        eD[atNo] = 1
                    else:
                        eD[atNo] += 1
            except Exception:
                pass
        return eD



##
# File:  OeDepictAlign.py
# Date:  27-Oct-2019  J. Westbrook
#
# Updates:
#
##
"""
Classes to depict aligned 2D chemical diagrams in a variety of media formats.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

# pylint: disable=too-many-lines

import logging
import os
import os.path

from openeye import oechem
from openeye import oedepict
from rcsb.utils.chem.OeCommonUtils import OeCommonUtils
from rcsb.utils.chem.OeDepict import OeDepictBase
from rcsb.utils.io.decorators import TimeoutException
from rcsb.utils.io.decorators import timeout
from rcsb.utils.io.FileUtil import FileUtil

logger = logging.getLogger(__name__)


class OeDepictAlignBase(OeDepictBase):
    """Base class for aligned 2D molecular renderings containing molecular
    object data and common display preferences.
    """

    def __init__(self):
        super(OeDepictAlignBase, self).__init__()
        #
        #
        self._refId = None
        self._refPath = None
        self._refMol = None
        self._refTitle = None
        self._refImagePath = None
        #
        self._fitId = None
        self._fitPath = None
        self._fitMol = None
        self._fitTitle = None
        self._fitImagePath = None
        #
        self._pairMolList = []
        #
        self._searchType = "default"
        self._minAtomMatchFraction = 0.50
        self._mcss = None
        self._miter = None
        #
        self._ss = None

    def setRefMol(self, oeMol, ccId, title=None, imagePath=None):
        try:
            self._refId = ccId
            self._refMol = oeMol
            #
            if title is not None:
                self._refMol.SetTitle(title)
                self._refTitle = title
            else:
                self._refMol.SetTitle(self._refId)
                self._refTitle = None
            #
            self._refImagePath = imagePath if imagePath is not None else self._refId + ".svg"
            return True
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return False

    def setFitMol(self, oeMol, ccId, title=None, imagePath=None):
        try:
            self._fitId = ccId
            self._fitMol = oeMol
            #
            if title is not None:
                self._fitMol.SetTitle(title)
                self._fitTitle = title
            else:
                self._fitMol.SetTitle(self._fitId)
                self._fitTitle = None
            #
            self._fitImagePath = imagePath if imagePath is not None else self._fitId + ".svg"
            return True
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return False

    def addFitMolList(self, oeMolList, suppressHydrogens=False, imageDirPath=".", imageFilePrefix=None):
        """Set the list of molecules to be compared with reference molecule by MCSS.

        From the input list build the internal pair list (self._pairMolList) of
        tuples  [(refId,refMol,refTitle,refImagePath,fitId,fitMol,fitTitle,fitImagePath),(),...]
        """
        self._pairMolList = []
        try:
            imageFilePrefix = imageFilePrefix + "-" if imageFilePrefix else ""
            for oeMol in oeMolList:
                refId = self._refId
                refImagePath = self._refImagePath
                refMol = oechem.OESuppressHydrogens(oechem.OEGraphMol(self._refMol)) if suppressHydrogens else self._refMol
                #
                fitId = str(oeMol.GetTitle()).upper()
                fitMol = oechem.OESuppressHydrogens(oechem.OEGraphMol(oeMol)) if suppressHydrogens else oeMol
                fitTitle = fitId + "/" + refId
                refTitle = refId + "/" + fitId
                fitImagePath = os.path.join(imageDirPath, imageFilePrefix + fitId + "-with-ref-" + refId + ".svg")
                self._pairMolList.append((refId, refMol, refTitle, refImagePath, fitId, fitMol, fitTitle, fitImagePath))
            return True
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return False

    def setPairMolList(self, pairOeMolList, suppressHydrogens=False, imageDirPath=".", imageFilePrefix=None):
        """Set the pairs of molecules to be compared by MCSS.

        From the input list build the internal pair list (self._pairMolList) of
        tuples  [(refId,refMol,refTitle,refImagePath,fitId,fitMol,fitTitle,fitImagePath),(),...]
        """
        self._pairMolList = []
        try:
            imageFilePrefix = imageFilePrefix + "-" if imageFilePrefix else ""
            for refId, refOeMol, fitId, fitOeMol in pairOeMolList:
                refImagePath = None
                refMol = oechem.OESuppressHydrogens(oechem.OEGraphMol(refOeMol)) if suppressHydrogens else refOeMol
                #
                fitMol = oechem.OESuppressHydrogens(oechem.OEGraphMol(fitOeMol)) if suppressHydrogens else fitOeMol
                fitTitle = fitId + "/" + refId
                refTitle = refId + "/" + fitId
                fitImagePath = os.path.join(imageDirPath, imageFilePrefix + "pair-fit-" + fitId + "-with-ref-" + refId + ".svg")
                refImagePath = os.path.join(imageDirPath, imageFilePrefix + "pair-fit-" + refId + "-with-ref-" + fitId + ".svg")
                self._pairMolList.append((refId, refMol, refTitle, refImagePath, fitId, fitMol, fitTitle, fitImagePath))
            return True
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return False

    def setSearchType(self, sType="default", minAtomMatchFraction=0.50):
        self._searchType = sType
        self._minAtomMatchFraction = minAtomMatchFraction
        return self._searchType

    def _setupMCSS(self, refmol):
        """Internal initialization for the MCSS comparison."""
        # self._mcss = oechem.OEMCSSearch(oechem.OEMCSType_Approximate)
        self._mcss = oechem.OEMCSSearch(oechem.OEMCSType_Exhaustive)
        #
        atomexpr, bondexpr = OeCommonUtils.getAtomBondExprOpts(self._searchType)

        self._mcss.Init(refmol, atomexpr, bondexpr)
        #
        # self._mcss.SetMCSFunc(OEMCSMaxBondsCompleteCycles())
        # self._mcss.SetMCSFunc(OEMCSMaxAtoms())
        #
        # Half of the reference molecule --
        #
        # nAtomsRef=refmol.NumAtoms()
        # self._mcss.SetMinAtoms(nAtomsRef/2)

    def _setupSubStructure(self, refmol):
        """Internal initialization for a substructure comparison."""

        #
        atomexpr, bondexpr = OeCommonUtils.getAtomBondExprOpts(self._searchType)
        #
        self._ss = oechem.OESubSearch(refmol, atomexpr, bondexpr)
        #

    def _setHighlightStyleRef(self, matchObj, refMol):
        refdisp = oedepict.OE2DMolDisplay(refMol, self._opts)
        #
        optInverseFit = False
        if self._params["highlightStyleRef"] == "ballAndStick":
            hstyle = oedepict.OEHighlightStyle_BallAndStick
        elif self._params["highlightStyleRef"] == "stick":
            hstyle = oedepict.OEHighlightStyle_Stick
        elif self._params["highlightStyleRef"] == "stickInverse":
            hstyle = oedepict.OEHighlightStyle_Stick
            optInverseFit = True
        elif self._params["highlightStyleRef"] == "ballAndStickInverse":
            hstyle = oedepict.OEHighlightStyle_BallAndStick
            optInverseFit = True
        else:
            hstyle = oedepict.OEHighlightStyle_BallAndStick

        if self._params["highLightMatchColorRef"] == "blue":
            myHighLightMatchColor = oechem.OEBlueTint
        elif self._params["highLightMatchColorRef"] == "green":
            myHighLightMatchColor = oechem.OEGreenTint
        elif self._params["highLightMatchColorRef"] == "pink":
            myHighLightMatchColor = oechem.OEPinkTint
        else:
            myHighLightMatchColor = oechem.OEBlueTint

        if self._params["highLightNotMatchColorRef"] == "blue":
            myHighLightNotMatchColor = oechem.OEBlueTint
        elif self._params["highLightNotMatchColorRef"] == "green":
            myHighLightNotMatchColor = oechem.OEGreenTint
        elif self._params["highLightNotMatchColorRef"] == "pink":
            myHighLightNotMatchColor = oechem.OEPinkTint
        else:
            myHighLightNotMatchColor = oechem.OEBlueTint

        #
        matchedatoms = oechem.OEIsAtomMember(matchObj.GetPatternAtoms())
        matchedbonds = oechem.OEIsBondMember(matchObj.GetPatternBonds())
        if optInverseFit:
            oedepict.OEAddHighlighting(refdisp, myHighLightNotMatchColor, hstyle, oechem.OENotAtom(matchedatoms), oechem.OENotBond(matchedbonds))
        else:
            oedepict.OEAddHighlighting(refdisp, myHighLightMatchColor, hstyle, matchedatoms, matchedbonds)

        return refdisp

    def _setHighlightStyleFit(self, matchObj, fitMol):
        fitdisp = oedepict.OE2DMolDisplay(fitMol, self._opts)

        optInverseFit = False
        if self._params["highlightStyleFit"] == "ballAndStick":
            hstyle = oedepict.OEHighlightStyle_BallAndStick
        elif self._params["highlightStyleFit"] == "stick":
            hstyle = oedepict.OEHighlightStyle_Stick
        elif self._params["highlightStyleFit"] == "stickInverse":
            hstyle = oedepict.OEHighlightStyle_Stick
            optInverseFit = True
        elif self._params["highlightStyleFit"] == "ballAndStickInverse":
            hstyle = oedepict.OEHighlightStyle_BallAndStick
            optInverseFit = True
        else:
            hstyle = oedepict.OEHighlightStyle_BallAndStick

        if self._params["highLightMatchColorFit"] == "blue":
            myHighLightMatchColor = oechem.OEBlueTint
        elif self._params["highLightMatchColorFit"] == "green":
            myHighLightMatchColor = oechem.OEGreenTint
        elif self._params["highLightMatchColorFit"] == "pink":
            myHighLightMatchColor = oechem.OEPinkTint
        else:
            myHighLightMatchColor = oechem.OEBlueTint

        if self._params["highLightNotMatchColorFit"] == "blue":
            myHighLightNotMatchColor = oechem.OEBlueTint
        elif self._params["highLightNotMatchColorFit"] == "green":
            myHighLightNotMatchColor = oechem.OEGreenTint
        elif self._params["highLightNotMatchColorFit"] == "pink":
            myHighLightNotMatchColor = oechem.OEPinkTint
        else:
            myHighLightNotMatchColor = oechem.OEBlueTint

        #
        # matchedatoms = oechem.OEIsAtomMember(matchObj.GetPatternAtoms())
        # matchedbonds = oechem.OEIsBondMember(matchObj.GetPatternBonds())
        matchedatoms = oechem.OEIsAtomMember(matchObj.GetTargetAtoms())
        matchedbonds = oechem.OEIsBondMember(matchObj.GetTargetBonds())

        if optInverseFit:
            oedepict.OEAddHighlighting(fitdisp, myHighLightNotMatchColor, hstyle, oechem.OENotAtom(matchedatoms), oechem.OENotBond(matchedbonds))
        else:
            oedepict.OEAddHighlighting(fitdisp, myHighLightMatchColor, hstyle, matchedatoms, matchedbonds)

        return fitdisp


class OeDepictMCSAlignMultiPage(OeDepictAlignBase):
    """Create 2D depictions of MCSS alignments. Inputs can be in the the form of pairs,
    lists, and pair lists of molecule object instances.

    Output images are rendered in a grid layout that can span multiple pages.
    """

    def __init__(self):
        super(OeDepictMCSAlignMultiPage, self).__init__()
        self.__grid = None
        self.__gridRows = None
        self.__gridCols = None
        self.__multi = None
        self.__image = None
        self.__citer = None
        #

    def alignPairListMulti(self, imagePath="multi.pdf"):
        aM = []
        self._params["gridCols"] = 2
        try:
            aM = self.__alignListMultiWorker(imagePath=imagePath, layout="pairs")
        except TimeoutException:
            logger.info("Timeout exception")
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return aM

    def alignOneWithListMulti(self, imagePath="multi.pdf"):
        aM = []
        try:
            aM = self.__alignListMultiWorker(imagePath=imagePath, layout="list")
        except TimeoutException:
            logger.info("Timeout exception")
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return aM

    def __setupImageMulti(self):
        """Internal method to configure a multipage image."""
        #
        self.__gridRows = self._params["gridRows"]
        self.__gridCols = self._params["gridCols"]
        #
        if self._params["pageOrientation"] == "landscape":
            self.__multi = oedepict.OEMultiPageImageFile(oedepict.OEPageOrientation_Landscape, oedepict.OEPageSize_US_Letter)
        else:
            self.__multi = oedepict.OEMultiPageImageFile(oedepict.OEPageOrientation_Portrait, oedepict.OEPageSize_US_Letter)

        self.__newPage()

    def __newPage(self):
        """Internal method to advance to a new page in a multipage configuration."""
        rows = self.__gridRows
        cols = self.__gridCols
        self.__image = self.__multi.NewPage()
        self.__grid = oedepict.OEImageGrid(self.__image, rows, cols)
        self.__grid.SetCellGap(self._params["cellGap"])
        self.__grid.SetMargins(self._params["cellMargin"])
        logger.debug("Num columns %d", self.__grid.NumCols())
        logger.debug("Num rows    %d", self.__grid.NumRows())
        self._opts = oedepict.OE2DMolDisplayOptions(self.__grid.GetCellWidth(), self.__grid.GetCellHeight(), oedepict.OEScale_AutoScale)
        self._assignDisplayOptions()
        self.__citer = self.__grid.GetCells()

    @timeout(500)
    def __alignListMultiWorker(self, imagePath="multi.pdf", layout="pairs"):
        """Working method comparing a reference molecule with a list of fit molecules.

        pairMolList = (refId,refMol,refTitle,refImagePath,fitId,fitMol,fitTitle,fitImagePath)

        Map of corresponding atoms is returned.

        Image Output is in multipage layout.
        """
        #
        self.__setupImageMulti()
        #
        atomMap = []
        firstOne = True
        iCount = 0
        for (refId, refMol, _, _, fitId, fitMol, fitTitle, _) in self._pairMolList:
            iCount += 1
            #
            oedepict.OEPrepareDepiction(refMol)
            self._setupMCSS(refMol)
            #
            #
            fitMol.SetTitle(fitTitle)
            #
            oedepict.OEPrepareDepiction(fitMol)
            #
            nAtomsRef = refMol.NumAtoms()
            nAtomsFit = fitMol.NumAtoms()
            minAtoms = min(nAtomsRef, nAtomsFit)
            mcssMinAtoms = int(minAtoms * self._minAtomMatchFraction)
            self._mcss.SetMinAtoms(mcssMinAtoms)

            # scaling
            refscale = oedepict.OEGetMoleculeScale(refMol, self._opts)
            fitscale = oedepict.OEGetMoleculeScale(refMol, self._opts)
            self._opts.SetScale(min(refscale, fitscale))

            unique = True
            self._miter = self._mcss.Match(fitMol, unique)
            logger.debug("mcss match completed for refId %s fitId %s", refId, fitId)
            if self._miter.IsValid():
                match = self._miter.Target()
                oedepict.OEPrepareAlignedDepiction(fitMol, self._mcss.GetPattern(), match)

                # Depict reference molecule with MCS highlighting
                if (firstOne and layout in ["list"]) or layout in ["pairs"]:
                    firstOne = False
                    if layout in ["pairs"]:
                        refdisp = self._setHighlightStyleRef(matchObj=match, refMol=self._mcss.GetPattern())
                    else:
                        refdisp = oedepict.OE2DMolDisplay(refMol, self._opts)
                    if not self.__citer.IsValid():
                        self.__newPage()
                    cell = self.__citer.Target()
                    self.__citer.Next()
                    oedepict.OERenderMolecule(cell, refdisp)
                    if self._params["cellBorders"]:
                        oedepict.OEDrawBorder(cell, oedepict.OEPen(oedepict.OEBlackPen))

                # Depict fit molecule with MCS highlighting
                fitdisp = self._setHighlightStyleFit(matchObj=match, fitMol=fitMol)
                if not self.__citer.IsValid():
                    self.__newPage()
                cell = self.__citer.Target()
                self.__citer.Next()
                oedepict.OERenderMolecule(cell, fitdisp)

                if self._params["cellBorders"]:
                    oedepict.OEDrawBorder(cell, oedepict.OEPen(oedepict.OEBlackPen))

                for mAt in match.GetAtoms():
                    atomMap.append((refId, mAt.pattern.GetName(), fitId, mAt.target.GetName()))

                logger.debug("mcss match completed for refId %s fitId %s total map length %d", refId, fitId, len(atomMap))

        logger.debug("writing image %s", imagePath)
        fU = FileUtil()
        fU.mkdirForFile(imagePath)
        oedepict.OEWriteMultiPageImage(imagePath, self.__multi)
        logger.debug("completed with map lenth %d", len(atomMap))
        #
        return atomMap


class OeDepictMCSAlignPage(OeDepictAlignBase):
    """Create 2D depictions of MCSS alignments. Inputs can be in the the form of pairs,
    lists, and pair lists of molecule object instances.

    Output images are rendered to a single page image with grid layout.
    """

    def __init__(self):
        super(OeDepictMCSAlignPage, self).__init__()
        self.__grid = None
        # self.__gridRows = None
        # self.__gridCols = None
        # self.__multi = None
        self.__image = None
        self.__citer = None
        #

    def __setupImage(self):
        """Internal method to configure a single pair alignment image."""
        #
        self.__image = oedepict.OEImage(self._params["imageSizeX"], self._params["imageSizeY"])
        rows = self._params["gridRows"]
        cols = self._params["gridCols"]
        self.__grid = oedepict.OEImageGrid(self.__image, rows, cols)
        self.__grid.SetCellGap(self._params["cellGap"])
        self.__grid.SetMargins(self._params["cellMargin"])
        logger.debug("Num columns %d", self.__grid.NumCols())
        logger.debug("Num rows    %d", self.__grid.NumRows())
        self._opts = oedepict.OE2DMolDisplayOptions(self.__grid.GetCellWidth(), self.__grid.GetCellHeight(), oedepict.OEScale_AutoScale)
        self._assignDisplayOptions()
        self.__citer = self.__grid.GetCells()

    def alignPair(self, imagePath="single-pair.png"):
        """Depict a single aligned ref/fit molecule pair or the first ref/fit molecule pair on the
        current _pairMolList.  Display options set for a single grid row with two columns.
        """
        self._params["gridCols"] = 2
        self._params["gridRows"] = 1
        self.__setupImage()
        #
        aM = []
        self._pairMolList = []
        self._pairMolList.append((self._refId, self._refMol, self._refTitle, None, self._fitId, self._fitMol, self._fitTitle, None))
        try:
            aM = self.__alignListWorker(imagePath=imagePath, layout="pairs")
        except TimeoutException:
            logger.info("Timeout exception")
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return aM

    def alignPairList(self, imagePath="single-pair-list.png"):
        aM = []
        self._params["gridCols"] = 2
        try:
            aM = self.__alignListWorker(imagePath=imagePath, layout="pairs")
        except TimeoutException:
            logger.info("Timeout exception")
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return aM

    def alignOneWithList(self, imagePath="single-list.png"):
        aM = []
        try:
            aM = self.__alignListWorker(imagePath=imagePath, layout="list")
        except TimeoutException:
            logger.info("Timeout exception")
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return aM

    # @timeout(15)
    def __alignListWorker(self, imagePath="single.pdf", layout="pairs"):
        """Working method comparing a reference molecule with a list of fit molecules.

        pairMolList = (refId,refMol,refTitle,refImgPath,fitId,fitMol,fitTitle,fitImgPath)

        Map of corresponding atoms is returned.

        Output image is a single-page with grid layout.
        """
        #
        self.__setupImage()
        #
        atomMap = []

        firstOne = True

        for (refId, refMol, _, _, fitId, fitMol, fitTitle, _) in self._pairMolList:
            #
            oedepict.OEPrepareDepiction(refMol)
            self._setupMCSS(refMol)
            #
            title = fitTitle if fitTitle else fitId
            fitMol.SetTitle(title)

            oedepict.OEPrepareDepiction(fitMol)
            #
            #
            nAtomsRef = refMol.NumAtoms()
            nAtomsFit = fitMol.NumAtoms()
            minAtoms = min(nAtomsRef, nAtomsFit)
            self._mcss.SetMinAtoms(int(minAtoms * self._minAtomMatchFraction))

            # scaling
            refscale = oedepict.OEGetMoleculeScale(refMol, self._opts)
            fitscale = oedepict.OEGetMoleculeScale(refMol, self._opts)
            self._opts.SetScale(min(refscale, fitscale))

            unique = True
            self._miter = self._mcss.Match(fitMol, unique)

            if self._miter.IsValid():
                match = self._miter.Target()
                oedepict.OEPrepareAlignedDepiction(fitMol, self._mcss.GetPattern(), match)

                # Depict reference molecule with MCS highlighting
                if (firstOne and layout in ["list"]) or layout in ["pairs"]:
                    firstOne = False
                    if layout in ["pairs"]:
                        refdisp = self._setHighlightStyleRef(matchObj=match, refMol=self._mcss.GetPattern())
                    else:
                        refdisp = oedepict.OE2DMolDisplay(refMol, self._opts)
                    if not self.__citer.IsValid():
                        break
                    cell = self.__citer.Target()
                    self.__citer.Next()
                    oedepict.OERenderMolecule(cell, refdisp)
                    if self._params["cellBorders"]:
                        oedepict.OEDrawBorder(cell, oedepict.OEPen(oedepict.OEBlackPen))

                # Depict fit molecule with MCS highlighting
                fitdisp = self._setHighlightStyleFit(matchObj=match, fitMol=fitMol)
                if not self.__citer.IsValid():
                    break
                cell = self.__citer.Target()
                self.__citer.Next()
                oedepict.OERenderMolecule(cell, fitdisp)

                if self._params["cellBorders"]:
                    oedepict.OEDrawBorder(cell, oedepict.OEPen(oedepict.OEBlackPen))

                for mAt in match.GetAtoms():
                    atomMap.append((refId, mAt.pattern.GetName(), fitId, mAt.target.GetName()))

        fU = FileUtil()
        fU.mkdirForFile(imagePath)
        oedepict.OEWriteImage(imagePath, self.__image)
        return atomMap


class OeDepictMCSAlign(OeDepictAlignBase):
    """Create 2D depictions of MCSS alignments. Inputs can be in the the form of pairs,
    lists, and pair lists of molecule object instances.

    Outputs are separate image files with a single diagram per file.
    """

    def __init__(self):
        super(OeDepictMCSAlign, self).__init__()
        #
        self.__imageRef = None
        self.__imageFit = None
        #

    def __setupImage(self):
        """Internal method to configure a single page image."""
        #
        self.__imageRef = oedepict.OEImage(self._params["imageSizeX"], self._params["imageSizeY"])
        self.__imageFit = oedepict.OEImage(self._params["imageSizeX"], self._params["imageSizeY"])
        self._opts = oedepict.OE2DMolDisplayOptions(self.__imageRef.GetWidth(), self.__imageRef.GetHeight(), oedepict.OEScale_AutoScale)
        self._assignDisplayOptions()

    def alignPair(self):
        """Depict a single aligned ref/fit molecule pair or the first ref/fit molecule pair on the
        current _pairMolList.  Display options set for a single grid row with two columns.
        """
        aM = []
        self._pairMolList = []
        self._pairMolList.append((self._refId, self._refMol, self._refTitle, self._refImagePath, self._fitId, self._fitMol, self._fitTitle, self._fitImagePath))
        try:
            aM = self.__alignListWorker(layout="pairs")
        except TimeoutException:
            logger.info("Timeout exception")
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return aM

    def alignPairList(self):
        aM = []
        try:
            aM = self.__alignListWorker(layout="pairs")
        except TimeoutException:
            logger.info("Timeout exception")
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return aM

    def alignOneWithList(self):
        aM = []
        try:
            aM = self.__alignListWorker(layout="list")
        except TimeoutException:
            logger.info("Timeout exception")
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return aM

    @timeout(15)
    def __alignListWorker(self, layout="pairs"):
        """Working method comparing a reference molecule with a list of fit molecules.

        pairMolList = (refId,refMol,refTitle,refImagePath, fitId,fitMol,fitTitle, fitImagePath)

        Map of corresponding atoms is returned.

        Writes separate output images for the reference and fit molecules for each comparison pair.
        """
        #
        atomMap = []
        firstOne = True
        fU = FileUtil()

        for (refId, refMol, _, refImagePath, fitId, fitMol, fitTitle, fitImagePath) in self._pairMolList:
            #
            self.__setupImage()
            #
            oedepict.OEPrepareDepiction(refMol)
            #
            self._setupMCSS(refMol)

            fitMol.SetTitle(fitTitle)
            oedepict.OEPrepareDepiction(fitMol)
            #
            nAtomsRef = refMol.NumAtoms()
            nAtomsFit = fitMol.NumAtoms()
            minAtoms = min(nAtomsRef, nAtomsFit)
            self._mcss.SetMinAtoms(int(minAtoms * self._minAtomMatchFraction))

            # scaling
            refscale = oedepict.OEGetMoleculeScale(refMol, self._opts)
            fitscale = oedepict.OEGetMoleculeScale(fitMol, self._opts)
            self._opts.SetScale(min(refscale, fitscale))

            unique = True
            self._miter = self._mcss.Match(fitMol, unique)
            if self._miter.IsValid():
                match = self._miter.Target()
                oedepict.OEPrepareAlignedDepiction(fitMol, self._mcss.GetPattern(), match)

                # Depict reference molecule with MCS highlighting
                if (firstOne and layout in ["list"]) or layout in ["pairs"]:
                    firstOne = False
                    if layout in ["pairs"]:
                        refdisp = self._setHighlightStyleRef(matchObj=match, refMol=self._mcss.GetPattern())
                    else:
                        refdisp = oedepict.OE2DMolDisplay(refMol, self._opts)
                    oedepict.OERenderMolecule(self.__imageRef, refdisp)
                    fU.mkdirForFile(refImagePath)
                    oedepict.OEWriteImage(refImagePath, self.__imageRef)

                # Depict fit molecule with MCS highlighting
                fitdisp = self._setHighlightStyleFit(matchObj=match, fitMol=fitMol)
                oedepict.OERenderMolecule(self.__imageFit, fitdisp)
                fU.mkdirForFile(fitImagePath)
                oedepict.OEWriteImage(fitImagePath, self.__imageFit)

                for mAt in match.GetAtoms():
                    atomMap.append((refId, mAt.pattern.GetName(), fitId, mAt.target.GetName()))

        return atomMap


class OeMCSAlignUtil(OeDepictAlignBase):
    """Perform MCSS alignments.  Inputs can be in the the form of pairs,
    lists, and pair lists of molecule object instances.
    """

    def __init__(self, maxMatches=2048):
        super(OeMCSAlignUtil, self).__init__()
        #
        self.__maxMatches = maxMatches

    def doAlign(self):
        """Test the MCSS comparison between the current reference and fit molecules -

        Return list of corresponding atoms on success or an empty list otherwise.
        """
        atomMap = []
        self._setupMCSS(self._refMol)
        #
        nAtomsRef = self._refMol.NumAtoms()
        nAtomsFit = self._fitMol.NumAtoms()
        minAtoms = min(nAtomsRef, nAtomsFit)
        self._mcss.SetMinAtoms(int(minAtoms * self._minAtomMatchFraction))
        unique = True
        self._miter = self._mcss.Match(self._fitMol, unique)
        if self._miter.IsValid():
            match = self._miter.Target()
            for mAt in match.GetAtoms():
                atomMap.append((self._refId, mAt.pattern.GetName(), self._fitId, mAt.target.GetName()))

        return atomMap

    def __getAtomIndex(self, oeMol):
        atD = {}
        try:
            for ii, atom in enumerate(oeMol.GetAtoms()):
                atD[atom.GetName().strip()] = ii
        except Exception:
            pass
        return atD

    def __getChargeIndex(self, oeMol):
        atD = {}
        try:
            for atom in oeMol.GetAtoms():
                atD[atom.GetName().strip()] = atom.GetFormalCharge()
        except Exception:
            pass
        return atD

    def __getNeighbors(self, oeMol, atNameList):
        nL = []
        try:
            for atom in oeMol.GetAtoms():
                aN = atom.GetName().strip()
                if aN in atNameList:
                    tL = []
                    for nbr in atom.GetAtoms():
                        tL.append(nbr.GetName())
                        nL.append((aN, tL))
        except Exception:
            pass
        return nL

    def doTestAlign(self):
        """Test the MCSS comparison between the current reference and fit molecules -

        Return list of corresponding atoms on success or an empty list otherwise.
        """
        self._minAtomMatchFraction = 0.9
        unique = True
        atomMap = []
        self._setupMCSS(self._refMol)
        #
        nAtomsRef = self._refMol.NumAtoms()
        nAtomsFit = self._fitMol.NumAtoms()
        #
        #
        minAtoms = min(nAtomsRef, nAtomsFit)
        self._mcss.SetMCSFunc(oechem.OEMCSMaxAtoms())
        self._mcss.SetMinAtoms(int(minAtoms * self._minAtomMatchFraction))
        unique = False

        self._miter = self._mcss.Match(self._fitMol, unique)
        if self._miter.IsValid():
            match = self._miter.Target()
            for mAt in match.GetAtoms():
                atomMap.append((self._refId, mAt.pattern.GetName(), self._fitId, mAt.target.GetName()))

        logger.debug("nAtomsRef %d nAtomsFit %d match length %d", nAtomsRef, nAtomsFit, len(atomMap))

        return atomMap

    def doAlignWithAnal(self):
        """Test the MCSS comparison between the current reference and fit molecules -

        Return list of corresponding atoms on success or an empty list otherwise.
        """
        self._minAtomMatchFraction = 0.9
        unique = True
        atomMap = []
        self._setupMCSS(self._refMol)
        #
        nAtomsRef = self._refMol.NumAtoms()
        nAtomsFit = self._fitMol.NumAtoms()
        #
        atomRefD = self.__getAtomIndex(self._refMol)
        atomFitD = self.__getAtomIndex(self._fitMol)
        #
        chgRefD = self.__getChargeIndex(self._refMol)
        chgFitD = self.__getChargeIndex(self._fitMol)
        #
        minAtoms = min(nAtomsRef, nAtomsFit)
        #
        self._mcss.SetMaxMatches(self.__maxMatches)
        nlim = oechem.OEGetMCSExhaustiveSearchTruncationLimit()
        logger.debug("search limit %d max matches %d", nlim, self.__maxMatches)
        self._mcss.SetMCSFunc(oechem.OEMCSMaxAtoms())
        self._mcss.SetMinAtoms(int(minAtoms * self._minAtomMatchFraction))
        unique = False

        atomRefMatchD = {}
        atomFitMatchD = {}
        self._miter = self._mcss.Match(self._fitMol, unique)
        if self._miter.IsValid():
            match = self._miter.Target()
            for mAt in match.GetAtoms():
                atomMap.append((self._refId, mAt.pattern.GetName(), self._fitId, mAt.target.GetName()))
                atomRefMatchD[mAt.pattern.GetName()] = mAt.target.GetName().strip()
                atomFitMatchD[mAt.target.GetName()] = mAt.pattern.GetName().strip()
        logger.debug("nAtomsRef %d nAtomsFit %d match length %d", nAtomsRef, nAtomsFit, len(atomMap))

        #
        # Get unmapped reference and fit atoms -
        #
        uRefAtomList = []
        uRefNList = []
        for atN in atomRefD:
            if atN not in atomRefMatchD:
                uRefAtomList.append(atN)
        #
        if len(uRefAtomList) > 0:
            uRefNList = self.__getNeighbors(self._refMol, uRefAtomList)
        #
        uFitAtomList = []
        uFitNList = []
        for atN in atomFitD:
            if atN not in atomFitMatchD:
                uFitAtomList.append(atN)
        #
        #
        if len(uFitAtomList) > 0:
            uFitNList = self.__getNeighbors(self._fitMol, uFitAtomList)
        #
        chgDifRefD = {}
        for atN0 in chgRefD:
            if atN0 in atomRefMatchD:
                atN = atomRefMatchD[atN0]
                # if ((atN not in chgFitD) or (chgRefD[atN0] != chgFitD[atN])):
                if chgRefD[atN0] != chgFitD[atN]:
                    chgDifRefD[atN] = chgRefD[atN0]
        #
        chgDifFitD = {}
        for atN0 in chgFitD:
            if atN0 in atomFitMatchD:
                atN = atomFitMatchD[atN0]
                # if ((atN not in chgRefD) or (chgFitD[atN0] != chgRefD[atN])):
                if chgFitD[atN0] != chgRefD[atN]:
                    chgDifFitD[atN] = chgFitD[atN0]

        return atomMap, uRefAtomList, uRefNList, chgDifRefD, uFitAtomList, uFitNList, chgDifFitD

    def doAlignList(self):
        """Test MCSS comparison between the current reference molecule with a list of fit molecules.

        pairMolList = (refId,refMol,refTitle,fitId,fitMol,fitTitle)

        Map of corresponding atoms is returned.

        """
        atomMap = []
        for (refId, refMol, _, _, fitId, fitMol, _, _) in self._pairMolList:
            #
            self._setupMCSS(refMol)
            #
            #
            nAtomsRef = refMol.NumAtoms()
            nAtomsFit = fitMol.NumAtoms()
            minAtoms = min(nAtomsRef, nAtomsFit)
            self._mcss.SetMinAtoms(int(minAtoms * self._minAtomMatchFraction))

            unique = True
            self._miter = self._mcss.Match(fitMol, unique)

            if self._miter.IsValid():
                match = self._miter.Target()
                for mAt in match.GetAtoms():
                    atomMap.append((refId, mAt.pattern.GetName(), fitId, mAt.target.GetName()))

        return atomMap


class OeSubStructureAlignUtil(OeDepictAlignBase):
    """Perform substructure alignments.  Inputs can be in the the form of pairs,
    lists, and pair lists of molecule object instances.
    """

    def __init__(self):
        super(OeSubStructureAlignUtil, self).__init__()
        #
        self.__unique = True

    def doAlign(self):
        """Test the MCSS comparison between the current reference and fit molecules -

        Return list of corresponding atoms on success or an empty list otherwise.
        """
        atomMap = []
        self._setupSubStructure(self._refMol)
        #
        unique = self.__unique
        self._miter = self._ss.Match(self._fitMol, unique)
        if self._miter.IsValid():
            match = self._miter.Target()
            for mAt in match.GetAtoms():
                atomMap.append((self._refId, mAt.pattern.GetName(), self._fitId, mAt.target.GetName()))

        return atomMap

    def doAlignList(self):
        """Test MCSS comparison between the current reference molecule with a list of fit molecules.

        pairMolList = (refId,refMol,refTitle,fitId,fitMol,fitTitle)

        Map of corresponding atoms is returned.

        """
        atomMap = []
        unique = self.__unique
        for (refId, refMol, _, _, fitId, fitMol, _, _) in self._pairMolList:
            #
            self._setupSubStructure(refMol)
            #
            self._miter = self._ss.Match(fitMol, unique)
            if self._miter.IsValid():
                match = self._miter.Target()
                for mAt in match.GetAtoms():
                    atomMap.append((refId, mAt.pattern.GetName(), fitId, mAt.target.GetName()))

        return atomMap


class OeDepictSubStructureAlign(OeDepictAlignBase):
    """Create 2D depictions of substructure alignments. Inputs can be in the the form of pairs,
    lists, and pair lists of molecule object instances.

    Outputs are separate image files with a single diagram per file.
    """

    def __init__(self):
        super(OeDepictSubStructureAlign, self).__init__()
        #
        self.__imageRef = None
        self.__imageFit = None
        #

    def __setupImage(self):
        """Internal method to configure a single page image."""
        #
        self.__imageRef = oedepict.OEImage(self._params["imageSizeX"], self._params["imageSizeY"])
        self.__imageFit = oedepict.OEImage(self._params["imageSizeX"], self._params["imageSizeY"])
        self._opts = oedepict.OE2DMolDisplayOptions(self.__imageRef.GetWidth(), self.__imageRef.GetHeight(), oedepict.OEScale_AutoScale)
        self._assignDisplayOptions()

    def alignPair(self):
        """Depict a single aligned ref/fit molecule pair or the first ref/fit molecule pair on the
        current _pairMolList.  Display options set for a single grid row with two columns.
        """
        aM = []
        self._pairMolList = []
        self._pairMolList.append((self._refId, self._refMol, self._refTitle, self._refImagePath, self._fitId, self._fitMol, self._fitTitle, self._fitImagePath))
        try:
            aM = self.__alignListWorker(layout="pairs")
        except TimeoutException:
            logger.info("Timeout exception")
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return aM

    def alignPairList(self):
        aM = []
        try:
            aM = self.__alignListWorker(layout="pairs")
        except TimeoutException:
            logger.info("Timeout exception")
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return aM

    def alignOneWithList(self):
        aM = []
        try:
            aM = self.__alignListWorker(layout="list")
        except TimeoutException:
            logger.info("Timeout exception")
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return aM

    @timeout(15)
    def __alignListWorker(self, layout="pairs"):
        """Working method comparing a reference molecule with a list of fit molecules.

        pairMolList = (refId,refMol,refTitle,refImagePath, fitId,fitMol,fitTitle, fitImagePath)

        Map of corresponding atoms is returned.

        Writes separate output images for the reference and fit molecules for each comparison pair.
        """
        #
        atomMap = []
        firstOne = True
        fU = FileUtil()
        for (refId, refMol, _, refImagePath, fitId, fitMol, fitTitle, fitImagePath) in self._pairMolList:
            #
            self.__setupImage()
            #
            oedepict.OEPrepareDepiction(refMol)
            #
            self._setupSubStructure(refMol)
            fitMol.SetTitle(fitTitle)
            oedepict.OEPrepareDepiction(fitMol)
            #
            # nAtomsRef = refMol.NumAtoms()
            # nAtomsFit = fitMol.NumAtoms()
            # minAtoms = min(nAtomsRef, nAtomsFit)
            # self._mcss.SetMinAtoms(int(minAtoms * self._minAtomMatchFraction))

            # scaling
            refscale = oedepict.OEGetMoleculeScale(refMol, self._opts)
            fitscale = oedepict.OEGetMoleculeScale(fitMol, self._opts)
            self._opts.SetScale(min(refscale, fitscale))
            #
            unique = True
            self._miter = self._ss.Match(fitMol, unique)
            if self._miter.IsValid():
                match = self._miter.Target()
                oedepict.OEPrepareAlignedDepiction(fitMol, self._ss.GetPattern(), match)

                # Depict reference molecule with MCS highlighting
                if (firstOne and layout in ["list"]) or layout in ["pairs"]:
                    firstOne = False
                    if layout in ["pairs"]:
                        refdisp = self._setHighlightStyleRef(matchObj=match, refMol=self._ss.GetPattern())
                    else:
                        refdisp = oedepict.OE2DMolDisplay(refMol, self._opts)
                    logger.info("writing %r", refImagePath)
                    oedepict.OERenderMolecule(self.__imageRef, refdisp)
                    fU.mkdirForFile(refImagePath)
                    oedepict.OEWriteImage(refImagePath, self.__imageRef)

                # Depict fit molecule with SS highlighting
                fitdisp = self._setHighlightStyleFit(matchObj=match, fitMol=fitMol)
                oedepict.OERenderMolecule(self.__imageFit, fitdisp)
                fU.mkdirForFile(fitImagePath)
                oedepict.OEWriteImage(fitImagePath, self.__imageFit)

                for mAt in match.GetAtoms():
                    atomMap.append((refId, mAt.pattern.GetName(), fitId, mAt.target.GetName()))

        return atomMap



##
# File:    ChemCompSearchIndexProvider.py
# Author:  J. Westbrook
# Date:    3-Mar-2020
#
# Updates:
#  13-Mar-2020 jdw Add formula index search method.
##
"""
Utilities to read and process an index of PDB chemical component definitions.
"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import logging
import os
import time
from collections import namedtuple

from rcsb.utils.chem.ChemCompMoleculeProvider import ChemCompMoleculeProvider
from rcsb.utils.chem.OeMoleculeFactory import OeMoleculeFactory
from rcsb.utils.io.IoUtil import getObjSize
from rcsb.utils.io.MarshalUtil import MarshalUtil

# from rcsb.utils.io.SingletonClass import SingletonClass
from rcsb.utils.multiproc.MultiProcUtil import MultiProcUtil

# from rcsb.utils.multiproc.MultiProcPoolUtil import MultiProcPoolUtil


logger = logging.getLogger(__name__)


MatchResults = namedtuple("MatchResults", "ccId oeMol searchType matchOpts screenType fpType fpScore oeIdx formula", defaults=(None,) * 9)


class ChemCompSearchIndexWorker(object):
    """A skeleton class that implements the interface expected by the multiprocessing
    for calculating search index candidates --
    """

    def __init__(self, ccObjD, **kwargs):
        self.__ccObjD = ccObjD
        _ = kwargs

    def buildRelatedList(self, dataList, procName, optionsD, workingDir):
        """Build search candidates for the input list of chemical component definitions
        and return index feature data.
        """
        _ = optionsD
        _ = workingDir
        # fetchLimit = optionsD.get("fetchLimit", None)
        limitPerceptions = optionsD.get("limitPerceptions", True)
        quietFlag = optionsD.get("quietFlag", False)
        successList = []
        failList = []
        retList = []
        diagList = []
        #
        try:
            retList, failList = self.__buildChemCompSearchIndex(procName, dataList, limitPerceptions=limitPerceptions, quietFlag=quietFlag)
            successList = sorted(set(dataList) - set(failList))
            if failList:
                logger.info("%s returns %d definitions with failures: %r", procName, len(failList), failList)

            logger.debug("%s built %d search candidates from %d definitions with failures %d", procName, len(retList), len(dataList), len(failList))
        except Exception as e:
            logger.exception("Failing %s for %d data items %s", procName, len(dataList), str(e))
        #
        return successList, retList, diagList

    def __buildChemCompSearchIndex(self, procName, ccIdList, limitPerceptions=False, quietFlag=False):
        """Internal method return a dictionary of extracted chemical component descriptors and formula."""
        rL = []
        fL = []
        try:
            for ccId in ccIdList:
                if ccId not in self.__ccObjD:
                    logger.error("%s missing chemical definition for %s", procName, ccId)
                    fL.append(ccId)
                    continue
                dataContainer = self.__ccObjD[ccId]
                # ----
                oemf = OeMoleculeFactory()
                if quietFlag:
                    oemf.setQuiet()
                tId = oemf.setChemCompDef(dataContainer)
                if tId != ccId:
                    logger.error("%s %s chemical component definition import error", procName, ccId)
                    fL.append(ccId)
                    continue
                relD = oemf.buildRelated(limitPerceptions=limitPerceptions)
                logger.debug("%s %s related molecular forms %d", procName, ccId, len(relD))
                if relD:
                    rL.extend([relD[v] for v in relD])
                else:
                    fL.append(ccId)
        except Exception as e:
            logger.exception("%s failing with %s", procName, str(e))
        return rL, fL


class ChemCompSearchIndexProvider(object):
    """Utilities to read and process the index of chemical component definitions search targets"""

    def __init__(self, **kwargs):
        #
        self.__cachePath = kwargs.get("cachePath", ".")
        self.__dirPath = os.path.join(self.__cachePath, "chem_comp")
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        self.__ccFileNamePrefix = kwargs.get("ccFileNamePrefix", "cc")
        self.__searchIdx = self.__reload(**kwargs)

    def testCache(self, minCount=None, logSizes=False):
        if logSizes and self.__searchIdx:
            logger.info("searchIdxD (%.2f MB)", getObjSize(self.__searchIdx) / 1000000.0)
        ok = self.__searchIdx and len(self.__searchIdx) >= minCount if minCount else self.__searchIdx is not None
        return ok

    def getIndex(self):
        return self.__searchIdx

    def getIndexEntry(self, searchCcId):
        try:
            return self.__searchIdx[searchCcId]
        except Exception as e:
            logger.debug("Get index entry %r failing with %s", searchCcId, str(e))
        return None

    def getIndexFilePath(self):
        return os.path.join(self.__dirPath, "%s-search-idx-chemical-components.json" % self.__ccFileNamePrefix)

    def __reload(self, **kwargs):
        """Reload or created index of PDB chemical components.

        Args:
            cachePath (str): path to the directory containing cache files
            ccIdxFileName (str): serialized chemical component data index file name

         Returns:
            (list): chemical component data containers
        """
        #
        searchIdxD = {}
        useCache = kwargs.get("useCache", True)
        molLimit = kwargs.get("molLimit", 0)
        numProc = kwargs.get("numProc", 1)
        maxChunkSize = kwargs.get("maxChunkSize", 20)
        limitPerceptions = kwargs.get("limitPerceptions", True)
        quietFlag = kwargs.get("quietFlag", True)
        searchIdxFilePath = self.getIndexFilePath()
        #
        if useCache and self.__mU.exists(searchIdxFilePath):
            _, fExt = os.path.splitext(searchIdxFilePath)
            searchIdxFormat = "json" if fExt == ".json" else "pickle"
            rdCcIdxD = self.__mU.doImport(searchIdxFilePath, fmt=searchIdxFormat)
            searchIdxD = {k: rdCcIdxD[k] for k in sorted(rdCcIdxD.keys())[:molLimit]} if molLimit else rdCcIdxD
        else:
            cmpKwargs = {k: v for k, v in kwargs.items() if k not in ["cachePath", "useCache", "molLimit"]}
            ccmP = ChemCompMoleculeProvider(cachePath=self.__cachePath, useCache=True, molLimit=molLimit, **cmpKwargs)
            ok = ccmP.testCache(minCount=molLimit, logSizes=True)
            if ok:
                searchIdxD = self.__updateChemCompSearchIndex(ccmP.getMolD(), searchIdxFilePath, molLimit, limitPerceptions, numProc, maxChunkSize, quietFlag)
                logger.info("Storing %s with data for %d search candidates (status=%r) ", searchIdxFilePath, len(searchIdxD), ok)
        #
        #
        for idxD in searchIdxD.values():
            idxD["atom-types"] = set(idxD["type-counts"].keys()) if "type-counts" in idxD else set()

        return searchIdxD

    def __updateChemCompSearchIndex(self, ccObjD, filePath, molLimit, limitPerceptions, numProc, maxChunkSize, quietFlag):
        searchIdxD = {}
        try:
            # Serialized index of chemical component search targets
            startTime = time.time()
            _, fExt = os.path.splitext(filePath)
            fileFormat = "json" if fExt == ".json" else "pickle"
            if numProc <= 1:
                searchIdxD = self.__buildChemCompSearchIndex(ccObjD, limitPerceptions=limitPerceptions, molLimit=molLimit)
            else:
                searchIdxD = self.__buildChemCompSearchIndexMulti(
                    ccObjD, limitPerceptions=limitPerceptions, molLimit=molLimit, numProc=numProc, maxChunkSize=maxChunkSize, quietFlag=quietFlag
                )

            ok = self.__mU.doExport(filePath, searchIdxD, fmt=fileFormat)
            endTime = time.time()
            logger.info("Storing %s (%s) with %d search definitions (status=%r) (%.4f seconds)", filePath, fileFormat, len(searchIdxD), ok, endTime - startTime)
        #
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        #
        return searchIdxD

    def __buildChemCompSearchIndex(self, ccObjD, limitPerceptions=False, molLimit=None):
        """Internal method return a dictionary of extracted chemical component descriptors and formula."""
        rD = {}
        try:
            for ii, ccId in enumerate(ccObjD, 1):
                if molLimit and ii > molLimit:
                    break
                # ----
                oemf = OeMoleculeFactory()
                oemf.setQuiet()
                tId = oemf.setChemCompDef(ccObjD[ccId])
                if tId != ccId:
                    logger.error("%s chemical component definition import error", ccId)
                smiD = oemf.buildRelated(limitPerceptions=limitPerceptions)
                logger.debug("%s related molecular forms %d", ccId, len(smiD))
                rD.update(smiD)
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return rD

    def __buildChemCompSearchIndexMulti(self, ccObjD, limitPerceptions=False, molLimit=None, numProc=2, maxChunkSize=20, quietFlag=False):
        #
        ccIdList = sorted(ccObjD.keys())[:molLimit] if molLimit else sorted(ccObjD.keys())
        logger.info("Input definition length %d numProc %d limitPerceptions %r", len(ccIdList), numProc, limitPerceptions)
        #
        rWorker = ChemCompSearchIndexWorker(ccObjD)
        # mpu = MultiProcPoolUtil(verbose=True)
        mpu = MultiProcUtil(verbose=True)
        optD = {"maxChunkSize": maxChunkSize, "limitPerceptions": limitPerceptions, "quietFlag": quietFlag}
        mpu.setOptions(optD)
        mpu.set(workerObj=rWorker, workerMethod="buildRelatedList")
        ok, failList, resultList, _ = mpu.runMulti(dataList=ccIdList, numProc=numProc, numResults=1, chunkSize=maxChunkSize)
        if failList:
            logger.info("Index definitions with failures (%d): %r", len(failList), failList)
        logger.info("Multi-proc status %r failures %r result length %r", ok, len(failList), len(resultList[0]))
        # JDW
        rD = {vD["name"]: vD for vD in resultList[0]}
        return rD

    def matchMolecularFormulaRange(self, typeRangeD, matchSubset=False):
        """Find matching formula for the input atom type range query (evaluates min <= ff <= max).

        Args:
            typeRangeD (dict): dictionary of element ranges {'<element_name>: {'min': <int>, 'max': <int>}}
            matchSubset (bool, optional): test for formula subset (default: False)

        Returns:
            (list):  chemical component identifiers with matching formula (MatchResults)
        """
        rL = []
        try:
            if not typeRangeD:
                return rL
            myTypeRangeD = {k.upper(): v for k, v in typeRangeD.items()}
            queryTypeS = set(myTypeRangeD.keys())
            for ccId, idxD in self.__searchIdx.items():
                tD = idxD["type-counts"]
                # targetTypeS = set(tD.keys())
                if not matchSubset and idxD["atom-types"] != queryTypeS:
                    continue
                #
                if not queryTypeS.issubset(idxD["atom-types"]):
                    continue
                match = True
                for atomType, rangeD in myTypeRangeD.items():
                    try:
                        if ("min" in rangeD and rangeD["min"] > tD[atomType]) or ("max" in rangeD and rangeD["max"] < tD[atomType]):
                            match = False
                            break
                    except Exception:
                        match = False
                        break
                if match:
                    # logger.info("%s formula %r query %r", ccId, idxD["type-counts"], typeRangeD)
                    rL.append(MatchResults(ccId=ccId, searchType="formula", formula=idxD["formula"]))
        except Exception as e:
            logger.exception("Failing for %r with %s", typeRangeD, str(e))
        return rL

    def filterMinimumMolecularFormula(self, typeCountD):
        """Find molecules with the minumum formula composition for the input atom type query (evaluates min <= ff).

        Args:
            typeCountD (dict): dictionary of element minimum values {'<element_name>: #}

        Returns:
            (list):  chemical component identifiers
        """
        rL = []
        try:
            if not typeCountD:
                return list(self.__searchIdx.keys())

            queryTypeS = set(typeCountD.keys())
            for ccId, idxD in self.__searchIdx.items():
                tD = idxD["type-counts"]
                if not queryTypeS.issubset(tD):
                    continue
                match = True
                for atomType, minCount in typeCountD.items():
                    try:
                        if minCount > tD[atomType]:
                            match = False
                            break
                    except Exception:
                        match = False
                        break
                if match:
                    rL.append(ccId)
        except Exception as e:
            logger.exception("Failing for %r with %s", typeCountD, str(e))
        return rL

    def filterMinimumFormulaAndFeatures(self, typeCountD, featureCountD):
        """Find molecules with the minumum formula and feature composition.

        Args:
            typeCountD (dict): dictionary of element minimum values {'<element_name>: #}
            featureCountD (dict): dictionary of feature minimum values {'<element_name>: #}

        Returns:
            (list):  chemical component identifiers
        """
        rL = []
        try:
            if not typeCountD or not featureCountD:
                return list(self.__searchIdx.keys())
            # ----
            featureQueryS = set(featureCountD.keys())
            typeQueryS = set(typeCountD.keys())
            #
            for ccId, idxD in self.__searchIdx.items():
                tD = idxD["type-counts"]
                fD = idxD["feature-counts"]
                #
                if not typeQueryS.issubset(tD) or not featureQueryS.issubset(fD):
                    continue

                match = True
                for atomType, minCount in typeCountD.items():
                    try:
                        if minCount > tD[atomType]:
                            match = False
                            break
                    except Exception:
                        match = False
                        break

                if not match:
                    continue
                #
                for featureType, minCount in featureCountD.items():
                    try:
                        if minCount > fD[featureType]:
                            match = False
                            break
                    except Exception:
                        match = False
                        break
                #
                if match:
                    rL.append(ccId)
        except Exception as e:
            logger.exception("Failing for %r with %s", typeCountD, str(e))
        return rL



##
# File:    OeMoleculeProvider.py
# Author:  J. Westbrook
# Date:    24-Oct-2019
#
# Updates:
#  28-Oct-2019 jdw incorporate all of the public Bird cc definitions
##
"""
Utilities deliver OE molecule data for PDB chemical component definitions
"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import logging
import os
import time

from rcsb.utils.chem.ChemCompMoleculeProvider import ChemCompMoleculeProvider
from rcsb.utils.chem.OeIoUtils import OeIoUtils
from rcsb.utils.io.MarshalUtil import MarshalUtil

# from rcsb.utils.io.SingletonClass import SingletonClass

logger = logging.getLogger(__name__)


class OeMoleculeProvider(object):
    """Utilities build and deliver OE molecule databases from PDB chemical component definition data"""

    def __init__(self, **kwargs):
        """Utilities build and deliver OE molecule databases from PDB chemical component definition data
        Args:
            cachePath (str, optional): path to the directory containing cache files (default: '.')
            molBuildType (str,optional): data source for building OE molecules (default: "model-xyz")
            oeFileNamePrefix (str, optional) file name prefix for all generated databases (default: "oe")

        """
        # Database file names with be prefixed with base prefix plus the molecular build type and perception options
        oeFileNamePrefixBase = kwargs.get("oeFileNamePrefix", "oe")
        limitPerceptions = kwargs.get("limitPerceptions", False)
        molBuildType = kwargs.get("molBuildType", "model-xyz")
        if limitPerceptions and molBuildType in ["oe-smiles", "oe-iso-smiles", "inchi"]:
            self.__oeFileNamePrefix = oeFileNamePrefixBase + "-" + molBuildType + "-limit"
        else:
            self.__oeFileNamePrefix = oeFileNamePrefixBase + "-" + molBuildType
        #
        cachePath = kwargs.get("cachePath", ".")
        self.__dirPath = os.path.join(cachePath, "oe_mol")
        #
        self.__fpDbD = {}
        self.__ssDb = None
        self.__oeMolD = {}
        self.__oeMolDb = None
        self.__oeMolDbTitleD = None
        #
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        self.__molCount = self.__reload(**kwargs)

    def testCache(self):
        return self.__mU.exists(os.path.join(self.__dirPath, self.__getOeMolFileName())) and self.__mU.exists(os.path.join(self.__dirPath, self.__getOeMolDbFileName()))

    def getSubSearchDb(self, screenType="SMARTS", numProc=1, forceRefresh=False):
        if not self.__ssDb or forceRefresh:
            oeIo = OeIoUtils()
            fp = os.path.join(self.__dirPath, self.__getSubSearchFileName(screenType))
            logger.info("Opening screened substructure search database %r", fp)
            self.__ssDb = oeIo.loadOeSubSearchDatabase(fp, screenType, numProc=numProc)
        return self.__ssDb

    def getFingerPrintDb(self, fpType, fpDbType="STANDARD", rebuild=False):
        if fpType not in self.__fpDbD or rebuild:
            oeIo = OeIoUtils()
            fastFpDbPath = os.path.join(self.__dirPath, self.__getFastFpDbFileName(fpType))
            oeMolDbFilePath = os.path.join(self.__dirPath, self.__getOeMolDbFileName())
            fpDb = oeIo.loadOeFingerPrintDatabase(oeMolDbFilePath, fastFpDbPath, inMemory=True, fpType=fpType, fpDbType=fpDbType)
            if fpDb:
                self.__fpDbD[fpType] = fpDb
        #
        return self.__fpDbD[fpType]

    def __getOeMolDbTitleIndex(self):
        oeMolDbTitleD = {}
        try:
            for idx in range(self.__oeMolDb.GetMaxMolIdx()):
                oeMolDbTitleD[self.__oeMolDb.GetTitle(idx)] = idx
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return oeMolDbTitleD

    def getOeMolDatabase(self):
        if not self.__oeMolDb:
            oeIo = OeIoUtils()
            self.__oeMolDb = oeIo.loadOeBinaryDatabaseAndIndex(os.path.join(self.__dirPath, self.__getOeMolDbFileName()))
            self.__oeMolDbTitleD = self.__getOeMolDbTitleIndex()
        return self.__oeMolDb, self.__oeMolDbTitleD

    def getOeMolD(self):
        try:
            if not self.__oeMolD:
                oeIo = OeIoUtils()
                self.__oeMolD = oeIo.readOeBinaryMolCache(os.path.join(self.__dirPath, self.__getOeMolFileName()))
                logger.info("Loading OE binary molecule cache length %d", len(self.__oeMolD))
            return self.__oeMolD
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return None

    def getMol(self, ccId):
        try:
            if not self.__oeMolD:
                oeIo = OeIoUtils()
                self.__oeMolD = oeIo.readOeBinaryMolCache(os.path.join(self.__dirPath, self.__getOeMolFileName()))
                logger.info("Loading OE binary molecule cache length %d", len(self.__oeMolD))
            return self.__oeMolD[ccId]
        except Exception as e:
            logger.exception("Get molecule %r failing with %s", ccId, str(e))
        return None

    def __getFastFpDbFileName(self, fpType):
        return "%s-fast-fp-database-%s.fpbin" % (self.__oeFileNamePrefix, fpType)

    def __getSubSearchFileName(self, screenType):
        return "%s-ss-database-%s.oeb" % (self.__oeFileNamePrefix, screenType)

    def __getOeMolDbFileName(self):
        return "%s-mol-db-components.oeb" % self.__oeFileNamePrefix

    def __getOeMolFileName(self):
        return "%s-mol-components.oeb" % self.__oeFileNamePrefix

    def __reload(self, **kwargs):
        """Reload the dictionary of OE molecules and related data artifacts for chemical component definitions.

        Args:
            molBuildType (str):  coordinates to use in building OE molecules from CIF components (model, ideal or None)
            limitPerceptions(bool): process input descriptors in essentially verbatim mode (default: True)
            fpTypeList (list): fingerprint type (TREE,PATH,MACCS,CIRCULAR,LINGO)
            screenTypeList (list): fast sub search screen type (MOLECULE, SMARTS, MDL, ... )
            useCache (bool, optional): flag to use cached files. Defaults to True.
            cachePath (str): path to the top cache directory. Defaults to '.'.
            numProc (int): number processors to engage in screen substructure search database generation.
            molLimit (int, optional): limiting number of molecules in data store (default: 0 no limit)
            suppressHydrogens (bool, optional): flag to suppress explicit hydrogens in the OE data store.

        Returns:
            (dict): dictionary of constructed OE molecules

        """
        useCache = kwargs.get("useCache", True)
        cachePath = kwargs.get("cachePath", ".")
        numProc = kwargs.get("numProc", 2)
        molLimit = kwargs.get("molLimit", 0)
        fpTypeList = kwargs.get("fpTypeList", ["TREE", "PATH", "MACCS", "CIRCULAR", "LINGO"])
        # screenTypeList = kwargs.get("screenTypeList", ["SMARTS"])
        screenTypeList = kwargs.get("screenTypeList", [])
        molBuildType = kwargs.get("molBuildType", "model-xyz")
        limitPerceptions = kwargs.get("limitPerceptions", False)
        quietFlag = kwargs.get("quietFlag", True)
        suppressHydrogens = kwargs.get("suppressHydrogens", False)
        logSizes = kwargs.get("logSizes", False)
        fpDbType = "STANDARD"
        #
        ccCount = 0
        oeCount = 0
        errCount = 0
        failIdList = []
        oeIo = OeIoUtils(quietFlag=quietFlag)
        # --------
        oeMolFilePath = os.path.join(self.__dirPath, self.__getOeMolFileName())
        if not useCache or (useCache and not self.__mU.exists(oeMolFilePath)):
            cmpKwargs = {k: v for k, v in kwargs.items() if k not in ["cachePath", "useCache", "molLimit"]}
            ccmP = ChemCompMoleculeProvider(cachePath=cachePath, useCache=True, molLimit=molLimit, **cmpKwargs)
            ok = ccmP.testCache(minCount=molLimit, logSizes=logSizes)
            ccObjD = ccmP.getMolD() if ok else {}
            ccCount = len(ccObjD)
            # -------
            startTime = time.time()
            oeCount, errCount, failIdList = oeIo.buildOeBinaryMolCache(
                oeMolFilePath, ccObjD, molBuildType=molBuildType, quietFlag=quietFlag, fpTypeList=fpTypeList, limitPerceptions=limitPerceptions, suppressHydrogens=suppressHydrogens
            )
            logger.info("Stored %d/%d OeMols (suppressH = %r) created with molBuildType %r (unconverted %d)", oeCount, ccCount, suppressHydrogens, molBuildType, errCount)
            if failIdList:
                logger.info("%r failures %r", molBuildType, failIdList)
            endTime = time.time()
            logger.info("Constructed %d/%d cached oeMols (%.4f seconds)", oeCount, ccCount, endTime - startTime)
        # --------
        oeMolDbFilePath = os.path.join(self.__dirPath, self.__getOeMolDbFileName())
        if not useCache or (useCache and not self.__mU.exists(oeMolDbFilePath)):
            startTime = time.time()
            molCount = oeIo.createOeBinaryDatabaseAndIndex(oeMolFilePath, oeMolDbFilePath)
            endTime = time.time()
            logger.info("Created and stored %d indexed OeMols in OE database format (%.4f seconds)", molCount, endTime - startTime)

        # --------
        if fpDbType == "FAST":
            for fpType in fpTypeList:
                startTime = time.time()
                #  Fast FP search database file names
                fpPath = os.path.join(self.__dirPath, self.__getFastFpDbFileName(fpType))
                if not useCache or (useCache and not self.__mU.exists(fpPath)):
                    ok = oeIo.createOeFingerPrintDatabase(oeMolDbFilePath, fpPath, fpType=fpType)
                    endTime = time.time()
                    logger.info("Created and stored %s fingerprint database (%.4f seconds)", fpType, endTime - startTime)
        # --------
        if molBuildType in ["oe-iso-smiles"]:
            for screenType in screenTypeList:
                startTime = time.time()
                fp = os.path.join(self.__dirPath, self.__getSubSearchFileName(screenType))
                if not useCache or (useCache and not self.__mU.exists(fp)):
                    ok = oeIo.createOeSubSearchDatabase(oeMolFilePath, fp, screenType=screenType, numProc=numProc)
                    endTime = time.time()
                    logger.info("Constructed screened substructure database (status %r) with screenType %s (%.4f seconds)", ok, screenType, endTime - startTime)
                    # ---------
                    ssDb = oeIo.loadOeSubSearchDatabase(fp, screenType=screenType, numProc=numProc)
                    ok = ssDb.NumMolecules() == oeCount
                    # ----------
        return oeCount
